from tala.model.stack import Stack
from tala import model
from tala.model import plan_item
from tala.model.semantic_object import SemanticObject
from tala.model.goal import Goal
from tala.utils.json_api import JSONAPIObject
from tala.model.action import Action
from tala.model.proposition import PredicateProposition


class UnableToDetermineOntologyException(Exception):
    pass


class Plan(Stack, SemanticObject):
    @classmethod
    def create_from_json_api_data(cls, plan_data, included):
        def generate_plan_items():
            if "plan_content" in plan_data["relationships"]:
                for plan_item_key in plan_data["relationships"]["plan_content"]["data"]:
                    plan_item_data = included.get_object_from_relationship(plan_item_key)
                    item = plan_item.PlanItem.create_from_json_api_data(plan_item_data, included)
                    if item is None:
                        raise BaseException("item is None", item)
                    yield item

        def generate_postplan_items():
            if "postplan_content" in plan_data["relationships"]:
                for plan_item_key in plan_data["relationships"]["postplan_content"]["data"]:
                    plan_item_data = included.get_object_from_relationship(plan_item_key)
                    item = plan_item.PlanItem.create_from_json_api_data(plan_item_data, included)
                    if item is None:
                        raise BaseException("item is None", item)
                    yield item

        goal_data = included.get_object_from_relationship(plan_data["relationships"]["goal"]["data"])
        goal = Goal.create_from_json_api_data(goal_data, included)
        preferred = False
        if "preferred" in plan_data["relationships"]:
            preference_data = included.get_object_from_relationship(plan_data["relationships"]["preferred"]["data"])
            preferred = PredicateProposition.create_from_json_api_data(preference_data, included)

        if not preferred and "preferred" in plan_data["attributes"]:
            preferred = True

        downdate_conditions = []
        if "downdate_conditions" in plan_data["relationships"]:
            for downdate_condition in plan_data["relationships"]["downdate_conditions"]["data"]:
                condition_data = included.get_object_from_relationship(downdate_condition)
                if downdate_condition["type"].endswith("PredicateProposition"):
                    downdate_conditions.append(PredicateProposition.create_from_json_api_data(condition_data, included))
                else:
                    downdate_conditions.append(model.condition.create_from_json_api_data(condition_data, included))
        superactions = []
        if "superactions" in plan_data["relationships"]:
            for action in plan_data["relationships"]["superactions"]["data"]:
                action_data = included.get_object_from_relationship(action)
                superactions.append(Action.create_from_json_api_data(action_data, included))

        accommodate_without_feedback = plan_data["attributes"].get("accommodate_without_feedback", False)

        unrestricted_accommodation = plan_data["attributes"].get("unrestricted_accommodation", False)
        reraise_on_resume = plan_data["attributes"].get("reraise_on_resume", True)
        max_answers = int(plan_data["attributes"].get("max_answers", -1))
        alternatives_predicate = plan_data["attributes"].get("alternatives_predicate", None)

        return (
            goal, Plan(list(generate_plan_items())), list(generate_postplan_items()), downdate_conditions, superactions,
            preferred, unrestricted_accommodation, reraise_on_resume, max_answers, alternatives_predicate,
            accommodate_without_feedback
        )

    def __init__(self, content=None):
        if content is None:
            content = set()
        Stack.__init__(self, content=content)
        SemanticObject.__init__(self)

    def is_ontology_specific(self):
        if not all([proposition.is_ontology_specific() for proposition in self.content]):
            return False
        if not self._ontology_names:
            return False
        return self._are_all_propositions_from_same_ontology()

    def _are_all_propositions_from_same_ontology(self):
        return len(self._ontology_names) < 2

    def as_json_api_dict(
        self, goal, preferred, accommodate_without_feedback, restart_on_completion, reraise_on_resume, max_answers,
        alternatives_predicate, postplan, downdate_conditions, superactions, unrestricted_accommodation
    ):
        def get_id():
            return f"{goal.ontology_name}:{goal.type_}:{goal.content}"

        dict_ = {
            "data": {
                "type": "tala.model.plan",
                "id": get_id(),
                "attributes": {
                    "version:id": "2",
                    "ontology_name": goal.ontology_name,
                    "accommodate_without_feedback": accommodate_without_feedback,
                    "restart_on_completion": restart_on_completion,
                    "reraise_on_resume": reraise_on_resume,
                    "max_answers": max_answers,
                    "alternatives_predicate": alternatives_predicate,
                    "unrestricted_accommodation": unrestricted_accommodation
                },
                "relationships": {}
            },
            "included": []
        }

        obj = JSONAPIObject.create_from_dict(dict_)
        if preferred is True:
            obj.add_attribute("preferred", True)
        elif preferred:
            obj.add_relationship("preferred", preferred.as_json_api_dict())
        obj.add_relationship("goal", goal.as_json_api_dict())

        for downdate_condition in downdate_conditions:
            obj.append_relationship("downdate_conditions", downdate_condition.as_json_api_dict())

        for item in reversed([plan_item_ for plan_item_ in self.iter_stack()]):
            obj.append_relationship("plan_content", item.as_json_api_dict())

        for plan_item_ in postplan:
            obj.append_relationship("postplan_content", plan_item_.as_json_api_dict())

        for superaction in superactions:
            obj.append_relationship("superactions", superaction.as_json_api_dict())

        return obj.as_dict

    @property
    def ontology_name(self):
        if not self._ontology_names:
            raise UnableToDetermineOntologyException("Expected semantic content but found none")

        if self.is_ontology_specific():
            return self._ontology_names.pop()

        message = "Expected all propositions %s\n\nin ontology %s\n\nbut they're from %s" % (
            self.content, self.content[0].ontology_name, self._ontology_names
        )
        raise UnableToDetermineOntologyException(message)

    @property
    def _ontology_names(self):
        ontology_names = set(proposition.ontology_name for proposition in self.content)
        return ontology_names

    def has_semantic_content(self):
        return True

    def iter_stack(self):
        return Stack.__iter__(self)

    def __iter__(self):
        flattened_items = self._flatten(self.content)
        return flattened_items.__iter__()

    def _flatten(self, items):
        result = []
        for item in items:
            if item.type_ == plan_item.TYPE_IF_THEN_ELSE:
                if item.consequent:
                    result.extend(self._flatten(item.get_consequent()))
                if item.alternative:
                    result.extend(self._flatten(item.get_alternative()))
            else:
                result.append(item)
        return result

    def remove(self, item_to_remove):
        items_to_remove = []
        for item in self.content:
            if item == item_to_remove:
                items_to_remove.append(item)
            elif item.type_ == plan_item.TYPE_IF_THEN_ELSE:
                self.remove_nested(item_to_remove, item)
        for item in items_to_remove:
            self.content.remove(item)

    def remove_nested(self, to_remove, item):
        if to_remove in item.consequent:
            item.consequent.remove(to_remove)
        if to_remove in item.alternative:
            item.alternative.remove(to_remove)
        for sub_item in item.alternative + item.consequent:
            if sub_item.type_ == plan_item.TYPE_IF_THEN_ELSE:
                self.remove_nested(to_remove, sub_item)

    def get_questions_in_plan_without_feature_question(self):
        for item in self:
            if item.type_ in plan_item.QUESTION_TYPES:
                question = item.content
                yield question

    def __str__(self):
        return "Plan(%s)" % self.content

    def __repr__(self):
        return "%s%s" % (self.__class__.__name__, (self.content))


class InvalidPlansException(Exception):
    pass



class ConfidenceEstimates():
    def __init__(
        self, perception_confidence=None, understanding_confidence=None, weighted_understanding_confidence=None
    ):
        self._perception_confidence = None
        self._understanding_confidence = None
        self._weighted_understanding_confidence = None
        if understanding_confidence is not None or perception_confidence is not None:
            self.set_realization_data(
                perception_confidence=perception_confidence, understanding_confidence=understanding_confidence
            )
            if weighted_understanding_confidence is not None:
                self.weighted_understanding_confidence = weighted_understanding_confidence

    @property
    def perception_confidence(self):
        return self._perception_confidence

    @property
    def understanding_confidence(self):
        return self._understanding_confidence

    @property
    def weighted_understanding_confidence(self):
        return self._weighted_understanding_confidence

    @weighted_understanding_confidence.setter
    def weighted_understanding_confidence(self, confidence):
        self._weighted_understanding_confidence = confidence

    @property
    def confidence(self):
        confidence_sources = [self.perception_confidence, self.understanding_confidence]
        if None in confidence_sources:
            return None
        return self.perception_confidence * self.understanding_confidence

    @property
    def weighted_confidence(self):
        confidence_sources = [self.perception_confidence, self.weighted_understanding_confidence]
        if None in confidence_sources:
            return None
        return self.perception_confidence * self.weighted_understanding_confidence

    def __eq__(self, other):
        return self.perception_confidence == other.perception_confidence \
            and self.understanding_confidence == other.understanding_confidence \
            and self.weighted_understanding_confidence == other.weighted_understanding_confidence

    def __str__(self):
        return f"ConfidenceEstimates(perception_confidence={self.perception_confidence}, " \
            + f"understanding_confidence={self.understanding_confidence}, " \
            + f"weighted_understanding_confidence={self.weighted_understanding_confidence})"

    def __repr__(self):
        return str(self)

    def set_realization_data(self, perception_confidence=None, understanding_confidence=None):
        self._perception_confidence = perception_confidence
        if self._perception_confidence is None:
            self._perception_confidence = 1.0
        self._understanding_confidence = understanding_confidence
        if self._understanding_confidence is None:
            self._understanding_confidence = 1.0
        self._weighted_understanding_confidence = self._understanding_confidence

    def uprank(self, amount):
        self._weighted_understanding_confidence *= (1 + amount)

    def downrank(self, amount):
        self._weighted_understanding_confidence *= (1 - amount)

    def build_string_from_attributes(self):
        string = ""
        if self.understanding_confidence is not None:
            string += ", understanding_confidence=%s" % self.understanding_confidence
        if self.weighted_understanding_confidence is not None:
            string += ", weighted_understanding_confidence=%s" % self.weighted_understanding_confidence
        if self.perception_confidence is not None:
            string += ", perception_confidence=%s" % self.perception_confidence
        return string



OFFER_ID = "OfferID"

OFFER_AUTHOR_ID = "UserID"
AUTHOR_ID = "AuthorUserID"

OFFER_QUOTA = "OfferQuota"
AUTHOR_QUOTA = "AuthorQuota"

NUM_CALLS = "NumCalls"
CALLS_LAST_PERIOD = "CallsLastPeriod"
ROW_KEY = "RowKey"

PARTITION_KEY = "PartitionKey"
HANDLER_DATA = "HandlerData"
BUDDY_GENERATOR_DATA = "BuddyGeneratorData"
USER_RATES = "UserRates"



from tala.model.polarity import Polarity


def facts_to_json_object(facts, session=None):
    def is_positive_predicate_proposition(proposition):
        return proposition.is_predicate_proposition() and proposition.get_polarity() == Polarity.POS

    return {
        proposition.predicate.get_name(): fact_to_json_object(proposition, session)
        for proposition in facts if is_positive_predicate_proposition(proposition)
    }


def fact_to_json_object(proposition, session=None):
    def get_grammar_entry(value):
        if "entities" in session:
            for entity in session["entities"]:
                if entity["name"] == value:
                    return entity["natural_language_form"]
        return None

    if proposition is None or proposition.individual is None:
        return None

    if session is None:
        session = {}

    resulting_dict = {
        "sort": proposition.predicate.sort.get_name(),
        "grammar_entry": get_grammar_entry(proposition.individual.value),
        "perception_confidence":
            proposition.confidence_estimates.perception_confidence if proposition.confidence_estimates else None,
        "understanding_confidence":
            proposition.confidence_estimates.understanding_confidence if proposition.confidence_estimates else None,
        "weighted_confidence":
            proposition.confidence_estimates.weighted_confidence if proposition.confidence_estimates else None,
        "weighted_understanding_confidence":
            proposition.confidence_estimates.weighted_understanding_confidence
            if proposition.confidence_estimates else None
    }  # yapf: disable

    resulting_dict.update(proposition.individual.value_as_json_object())

    return resulting_dict



import os

from tala.config import DddConfig, DddConfigNotFoundException
from tala.ddd.loading.ddd_loader import DDDLoader
from tala.utils.chdir import chdir


class DddNotFoundException(Exception):
    pass


class OverriddenDddConfigNotFoundException(Exception):
    pass


class DDDSetLoader(object):
    def __init__(self, overridden_ddd_config_paths=None):
        self._overridden_ddd_configs = overridden_ddd_config_paths or []
        self._validate_that_ddds_in_overridden_ddd_configs_exist()
        self._validate_that_configs_in_overridden_ddd_configs_exist()

    def _validate_that_ddds_in_overridden_ddd_configs_exist(self):
        for overridden_config in self._overridden_ddd_configs:
            if not os.path.exists(overridden_config.ddd_name):
                raise DddNotFoundException(
                    "Expected overridden DDD '%s' to exist in the working directory '%s', but it doesn't." %
                    (overridden_config.ddd_name, os.getcwd())
                )

    def _validate_that_configs_in_overridden_ddd_configs_exist(self):
        for overridden_config in self._overridden_ddd_configs:
            with chdir(overridden_config.ddd_name):
                try:
                    DddConfig(overridden_config.path).read()
                except DddConfigNotFoundException:
                    raise OverriddenDddConfigNotFoundException(
                        "Expected DDD config '%s' to exist in DDD '%s' but it was not found." %
                        (overridden_config.path, overridden_config.ddd_name)
                    )

    def ddds_as_list(self, ddds, path=".", *args, **kwargs):
        with chdir(path):
            return list(self._load_ddds(ddds, *args, **kwargs))

    def _load_ddds(self, ddd_names, *args, **kwargs):
        configs = {}
        for ddd_name in ddd_names:
            configs[ddd_name] = self._ddd_config(ddd_name)

        for ddd_name in ddd_names:
            config = configs[ddd_name]
            yield self._load_ddd(ddd_name, config, *args, **kwargs)

    def _load_ddd(self, *args, **kwargs):
        ddd_loader = DDDLoader(*args, **kwargs)
        return ddd_loader.load()

    def _ddd_config(self, ddd_name):
        for overridden_config in self._overridden_ddd_configs:
            if overridden_config.ddd_name == ddd_name:
                with chdir(ddd_name):
                    return DddConfig(overridden_config.path).read()
        with chdir(ddd_name):
            return DddConfig().read()



import unittest

from tala.model.service_action_outcome import SuccessfulServiceAction, FailedServiceAction


class ServiceActionOutcomeTests(unittest.TestCase):
    def test_successful_service_action_is_successful(self):
        self.given_created_successful_service_action()
        self.when_is_successful_is_retrieved()
        self.then_result_is(True)

    def given_created_successful_service_action(self):
        self._service_action_outcome = SuccessfulServiceAction()

    def when_is_successful_is_retrieved(self):
        self._actual = self._service_action_outcome.is_successful

    def then_result_is(self, expected):
        self.assertEqual(expected, self._actual)

    def test_failed_service_action_is_not_successful(self):
        self.given_created_failed_service_action()
        self.when_is_successful_is_retrieved()
        self.then_result_is(False)

    def given_created_failed_service_action(self, failure_reason="mock_failure_reason"):
        self._service_action_outcome = FailedServiceAction(failure_reason)

    def test_failure_reason(self):
        self.given_created_failed_service_action(failure_reason="mock_failure_reason")
        self.when_failure_reason_is_retrieved()
        self.then_result_is("mock_failure_reason")

    def when_failure_reason_is_retrieved(self):
        self._actual = self._service_action_outcome.failure_reason

    def test_successful_service_action_equals_successful_service_action(self):
        self.assertEqual(SuccessfulServiceAction(), SuccessfulServiceAction())

    def test_successful_service_action_not_equals_non_service_action_outcome(self):
        self.assertNotEqual(SuccessfulServiceAction(), "non_service_action_outcome")

    def test_failed_service_action_equals_failed_service_action_with_same_failure_reason(self):
        self.assertEqual(FailedServiceAction("mock_failure_reason"), FailedServiceAction("mock_failure_reason"))

    def test_failed_service_action_not_equals_failed_service_action_with_other_failure_reason(self):
        self.assertNotEqual(FailedServiceAction("mock_failure_reason_1"), FailedServiceAction("mock_failure_reason_2"))

    def test_failed_service_action_not_equals_non_service_action_outcome(self):
        self.assertNotEqual(FailedServiceAction("mock_failure_reason"), "non_service_action_outcome")

    def test_successful_service_action_not_equals_failed_service_action(self):
        self.assertNotEqual(SuccessfulServiceAction(), FailedServiceAction("mock_failure_reason"))

    def test_successful_service_action_repr(self):
        self.assertEqual("SuccessfulServiceAction()", repr(SuccessfulServiceAction()))

    def test_successful_service_action_str(self):
        self.assertEqual("SuccessfulServiceAction()", str(SuccessfulServiceAction()))

    def test_failed_service_action_repr(self):
        self.assertEqual("FailedServiceAction('mock_failure_reason')", repr(FailedServiceAction("mock_failure_reason")))

    def test_failed_service_action_str(self):
        self.assertEqual("FailedServiceAction(mock_failure_reason)", str(FailedServiceAction("mock_failure_reason")))



from datetime import datetime, timedelta
import time

import requests
from requests.exceptions import ConnectionError, ConnectTimeout


def await_endpoint(url, timeout=None):
    def connect():
        response = requests.get(url, timeout=timeout / 10)
        response.raise_for_status()

    def is_responding():
        try:
            connect()
            return True
        except (ConnectionError, ConnectTimeout):
            return False

    start_time = datetime.now()
    end_time = start_time + timedelta(seconds=timeout)
    while datetime.now() < end_time:
        if is_responding():
            return
        time.sleep(0.1)
        continue
    connect()



class Modality:
    TEXT = "text"
    SPEECH = "speech"
    HAPTIC = "haptic"
    OTHER = "other"
    REQUIRES_UTTERANCE = [SPEECH, TEXT]
    ALLOWS_UTTERANCE = [SPEECH, TEXT, OTHER, HAPTIC]
    SUPPORTED_MODALITIES = [SPEECH, TEXT, HAPTIC, OTHER]
    MODALITIES = [None] + SUPPORTED_MODALITIES



TYPE = "type"
ATTRIBUTES = "attributes"
MATCH = "match"
TOKEN = "token"
REPLACEMENT = "replacement"
CORRECTION_ENTRY_TYPE = "correction_entry"
ATTRIBUTE_LIST = ["ddd_name", "voice", "locale"]


def is_correction_entry_matching(entry, token):
    return entry[TYPE] == CORRECTION_ENTRY_TYPE and entry[ATTRIBUTES][MATCH] == token


def entry_matches_input(dict_attributes, input_entry):
    def attribute_is_required(attribute):
        if attribute in dict_attributes and dict_attributes.get(attribute):
            return True
        return False

    for attribute in ATTRIBUTE_LIST:
        if attribute_is_required(attribute):
            if dict_attributes[attribute] != input_entry.get(attribute):
                return False
    return True


class Lexicon:
    def __init__(self, entries):
        self._entries = entries

    def generate_pronunciation_text(self, input_entry):
        for entry in self._entries:
            if is_correction_entry_matching(entry, input_entry[TOKEN]):
                if entry_matches_input(entry[ATTRIBUTES], input_entry):
                    return entry[ATTRIBUTES][REPLACEMENT]
        return input_entry[TOKEN]

    def _create_replacements(self, input_entry):
        for entry in self._entries:
            if entry_matches_input(entry[ATTRIBUTES], input_entry):
                yield entry[ATTRIBUTES][MATCH], entry[ATTRIBUTES][REPLACEMENT]

    def get_replacement_dict(self, input_entry):
        return {k: v for k, v in self._create_replacements(input_entry)}



import contextlib
import os


class DirectoryException(Exception):
    pass


@contextlib.contextmanager
def chdir(new_dir):
    working_dir = os.getcwd()
    try:
        os.chdir(str(new_dir))
    except OSError:
        raise DirectoryException("Could not change directory from %r to %r" % (working_dir, new_dir))
    try:
        yield
    finally:
        os.chdir(working_dir)



import json
import requests
import threading
import time

TIMEOUT = 2.0


class StreamListener(threading.Thread):
    LEGACY_STREAMER_BASE_URL = 'https://tala-event-sse.azurewebsites.net/event-sse'
    STREAMER_BASE_URL = 'https://tala-sse-ng-g6bpb0cncyc4htg3.swedencentral-01.azurewebsites.net/event-sse'

    def __init__(self, streamer_url, session_id, logger):
        self._streamer_url = streamer_url
        self.current_data = None
        self.current_event = None
        self.stream_started = threading.Event()
        self._please_stop = threading.Event()
        self._stopped = threading.Event()
        self._session_id = session_id
        self._streaming_started = None
        self._streaming_ended = None
        self._logger = logger
        super().__init__(daemon=True)

    def run(self):
        self._streamed_messages = []
        response = requests.get(f"{self._streamer_url}/{self._session_id}", stream=True)
        self.stream_started.set()
        self._logger.info("listening to stream")
        for line in response.iter_lines():
            if line:  # filter out keep-alive new lines
                decoded_line = line.decode('utf-8')
                self.process_line(decoded_line)
            if self._please_stop.is_set():
                break
        self._logger.info("stream listener stopped listening", received_messages=self._streamed_messages)
        self._stopped.set()

    def process_line(self, line):
        def is_event():
            return line.startswith("event: ")

        def is_data():
            return line.startswith("data: ")

        def extract_event():
            return line[len("event: "):]

        def extract_data():
            return line[len("data: "):]

        self._logger.info("stream listener processing line", line=line)
        if is_event():
            self.current_event = extract_event()
        if is_data() and self.current_event:
            self.current_data = extract_data()
            self.create_event_and_store()

    def create_event_and_store(self):
        try:
            if self.current_event == "STREAMING_DONE":
                json_event = f'{{"event": "{self.current_event}"}}'
                self._streaming_ended = time.time()
                self.please_stop()
            if self.current_event == "STREAMING_CHUNK":
                if not bool(self._streaming_started):
                    self._streaming_started = time.time()

                json_event = f'{{"event": "{self.current_event}", "data": "{self.current_data}"}}'
            if json_event:
                self._streamed_messages.append(json.loads(json_event))
        except UnboundLocalError:
            pass
        self.current_event = None

    def please_stop(self):
        self._please_stop.set()

    @property
    def payload(self):
        self._stopped.wait(TIMEOUT)
        return self._streamed_messages

    @property
    def system_utterance(self):
        utterance = ""
        for item in self.payload:
            try:
                utterance += item["data"]
            except KeyError:
                pass
        return utterance.strip()

    @property
    def streaming_started(self):
        return self._streaming_started

    @property
    def streaming_ended(self):
        return self._streaming_ended



from tala.utils.as_json import AsJSONMixin
from tala.utils.json_api import JSONAPIMixin


class ExtendedDDD(AsJSONMixin, JSONAPIMixin):
    """ This is in practice a DDD with the addition of a parameter retriever and a parser """
    def __init__(self, ddd, parameter_retriever, parser, path=None):
        super(ExtendedDDD, self).__init__()
        self._ddd = ddd
        self._parameter_retriever = parameter_retriever
        self._parser = parser
        self._path = path

    @property
    def parameter_retriever(self):
        return self._parameter_retriever

    @property
    def parser(self):
        return self._parser

    @property
    def path(self):
        return self._path

    @property
    def ddd(self):
        return self._ddd

    @property
    def name(self):
        return self.ddd.name

    @property
    def ontology(self):
        return self.ddd.ontology

    @property
    def domain(self):
        return self.ddd.domain

    @property
    def service_interface(self):
        return self.ddd.service_interface

    @property
    def language_codes(self):
        return self.ddd.language_codes

    def reset(self):
        self.ontology.reset()
        if self.parser is not None:
            self.parser.clear()

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, (self.ddd, self.parameter_retriever, self.parser))

    def as_dict(self):
        return self.ddd.as_dict()

    def ensure_relevant_entities_added(self, entities):
        self.add_relevant_entities(entities)

    def add_relevant_entities(self, entities):
        for entity in entities:
            self._ensure_relevant_entity_added(entity)

    def _ensure_relevant_entity_added(self, entity):
        def entity_is_relevant():
            if not entity.ddd_name or self.name == entity.ddd_name:
                return self.ontology.has_sort(entity.sort)
            return False

        def is_entity_dynamic():
            sort = self.ontology.get_sort(entity.sort)
            return sort.is_dynamic()

        if entity_is_relevant() and is_entity_dynamic():
            self.ontology.ensure_individual_exists(entity.name, entity.sort)



def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    """
    For float comparison. See https://www.python.org/dev/peps/pep-0485/#proposed-implementation.

    Use relative tolerance, absolute tolerance, or both.

    :param a: Left operand
    :param b: Right operand
    :param rel_tol: Relative tolerance. The amount of error allowed. E.g. 5%, or the default: 1 PPB.
    :param abs_tol: Minimum absolute tolerance level.
    :return: True if a and b are within at least one of the relative or the absolute tolerances. False otherwise.
    """
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)



import datetime
import logging.config
import os

from structlog.processors import JSONRenderer, TimeStamper, StackInfoRenderer, format_exc_info, \
    ExceptionPrettyPrinter, KeyValueRenderer
from structlog.stdlib import add_log_level, filter_by_level, add_logger_name
import structlog

DEFAULT_LOG_LEVEL = logging.WARNING
TESTING_PROCESSORS = [
    filter_by_level,
    add_log_level,
    add_logger_name,
    TimeStamper(fmt="iso", key="time"),
    StackInfoRenderer(),
    format_exc_info,
    ExceptionPrettyPrinter(),
    KeyValueRenderer(sort_keys=True, repr_native_str=False),
]
STDOUT_HANDLER = {
    "stdout": {
        "class": "logging.StreamHandler",
        "stream": "ext://sys.stdout",
        "formatter": "plain",
    }
}


def get_logger():
    return structlog.get_logger("tdm")


def configure_file_logging(file_basename=None, level=None):
    if file_basename is None:
        file_basename = _log_basename("tdm")
    _ensure_log_folder_exists()
    filename = "logs/%s" % file_basename
    handlers = {
        "file": {
            "class": "logging.handlers.WatchedFileHandler",
            "filename": filename,
            "formatter": "plain",
            "mode": "w",
        }
    }
    return _configure_logging(handlers, level)


def configure_stdout_logging(level=None):
    return _configure_logging(STDOUT_HANDLER, level)


def configure_test_logging(level=None):
    level = level or logging.INFO
    return _configure_logging(STDOUT_HANDLER, level, processors=TESTING_PROCESSORS)


def _configure_logging(handlers, level=None, processors=None):
    level = level or DEFAULT_LOG_LEVEL
    logging.config.dictConfig({
        "version": 1,
        "formatters": {
            "plain": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.dev.ConsoleRenderer(colors=False),
            },
        },
        "handlers": handlers,
        "loggers": {
            "": {
                "handlers": list(handlers.keys()),
                "level": level,
                "propagate": False,
            },
        }
    })
    default_processors = [
        filter_by_level,
        add_log_level,
        add_logger_name,
        TimeStamper(fmt="iso", key="time"),
        StackInfoRenderer(),
        format_exc_info,
        JSONRenderer(sort_keys=True),
    ]
    processors = processors or default_processors
    structlog.configure(
        processors=processors,
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def _log_basename(prefix):
    now = datetime.datetime.now()
    timestamp = now.strftime("%Y-%m-%dT%H-%M-%S")
    return "%s_%s.log" % (prefix, timestamp)


def _ensure_log_folder_exists():
    try:
        os.mkdir("logs")
    except OSError:
        pass


def configure_and_get_test_logger():
    configure_file_logging("test_suite.log", level=logging.DEBUG)
    return structlog.get_logger("test_suite")



import os

from tala.utils.chdir import chdir
from tala.ddd.loading.ddd_loader import DDDLoader
from tala.ddd.parser import Parser

from tala.ddd.extended_ddd import ExtendedDDD
from tala.ddd.services.parameters.retriever import ParameterRetriever


class ExtendedDDDLoader(object):
    def __init__(self, ddd_manager, name, ddd_config, rerank_amount):
        ddd_loader = DDDLoader(name, ddd_config)
        self._ddd = ddd_loader.load()
        self._ddd_config = ddd_config
        self._ddd_manager = ddd_manager
        self._rerank_amount = rerank_amount

    def load(self):
        path = os.path.join(os.getcwd(), self._ddd.name)

        with chdir(self._ddd.name):
            parser = Parser(self._ddd.name, self._ddd.ontology, self._ddd.domain.name)
            parameter_retriever = ParameterRetriever(self._ddd.service_interface, self._ddd.ontology)

        ddd = ExtendedDDD(self._ddd, parameter_retriever, parser, path)
        return ddd



from tala.utils.as_json import AsJSONMixin
from tala.utils.equality import EqualityMixin
from tala.utils.json_api import JSONAPIMixin, IncludedObject
from tala import model
from tala.ddd.services.service_interface import ServiceInterface


class DDD(AsJSONMixin, EqualityMixin, JSONAPIMixin):
    @classmethod
    def create_from_json_api_data(cls, ddd_as_dict):
        ddd_entry = ddd_as_dict["data"]
        included = IncludedObject(ddd_as_dict["included"])
        name = ddd_entry["attributes"]["name"]
        path = ddd_entry["attributes"]["path"]

        ontology_object = included.get_object_from_relationship(ddd_entry["relationships"]["ontology"]["data"])
        domain_object = included.get_object_from_relationship(ddd_entry["relationships"]["domain"]["data"])
        service_interface_object = included.get_object_from_relationship(
            ddd_entry["relationships"]["service_interface"]["data"]
        )
        ontology = model.ontology.Ontology.create_from_json_api_data(ontology_object, included)
        domain = model.domain.Domain.create_from_json_api_data(domain_object, included, ontology)
        service_interface = ServiceInterface.create_from_json_api_data(service_interface_object, included)

        return cls(name, ontology, domain, service_interface, path)

    def __init__(self, name, ontology, domain, service_interface, path=None):
        super(DDD, self).__init__()
        self._name = name
        self._ontology = ontology
        self._domain = domain
        self._service_interface = service_interface
        self._path = path

    @property
    def name(self):
        return self._name

    @property
    def ontology(self):
        return self._ontology

    @property
    def domain(self):
        return self._domain

    @property
    def service_interface(self):
        return self._service_interface

    @property
    def path(self):
        return self._path

    def __repr__(self):
        return "%s%s" % (self.__class__.__name__, (self.name, self.ontology, self.domain, self.service_interface))

    def as_dict(self):
        return {
            "ddd_name": self._name,
            "ontology": self.ontology,
            "domain": self.domain,
            "service_interface": self.service_interface,
        }

    @property
    def json_api_id(self):
        return f"{self.name}"

    @property
    def json_api_attributes(self):
        return ["name", "path"]

    @property
    def json_api_relationships(self):
        return ["ontology", "domain", "service_interface"]



from io import StringIO
import os
import re

from tala.config import BackendConfig, DddConfig
from tala.ddd.maker import utils
from tala.utils import chdir


class UnexpectedCharactersException(Exception):
    pass


class DddMaker(object):
    def __init__(self, ddd_name, use_rgl, target_dir="."):
        self._validate(ddd_name)
        self._ddd_name = ddd_name
        self._use_rgl = use_rgl
        self._class_name_prefix = self.directory_to_class_name(ddd_name)
        self._target_dir = target_dir

    @staticmethod
    def directory_to_class_name(directory_name):
        name_with_capitalized_words = directory_name.title()
        class_name = re.sub("[_]", "", name_with_capitalized_words)
        return class_name

    @staticmethod
    def _validate(name):
        if re.match(r'^[0-9a-zA-Z_]+$', name) is None:
            raise UnexpectedCharactersException(
                f"Expected only alphanumeric ASCII and underscore characters in DDD name '{name}', but found others."
            )

    def make(self):
        self._ensure_target_dir_exists()
        self._create_ddd_module()
        self._create_domain_skeleton_file()
        self._create_ontology_skeleton_file()
        self._create_service_interface_skeleton_file()
        self._create_configs()
        self._create_interaction_tests()

    def _ensure_target_dir_exists(self):
        if not os.path.exists(self._target_dir):
            os.mkdir(self._target_dir)

    def _create_ddd_module(self):
        ddd_path = self._ddd_path()
        if not os.path.exists(ddd_path):
            os.makedirs(ddd_path)
        self._create_empty_file("__init__.py")

    def _create_configs(self):
        with chdir.chdir(self._target_dir):
            BackendConfig.write_default_config(ddd_name=self._ddd_name)
            with chdir.chdir(self._ddd_name):
                DddConfig.write_default_config(use_rgl=self._use_rgl)

    def _create_ontology_skeleton_file(self):
        self._create_skeleton_file("ontology_template.xml", "ontology.xml")

    def _create_domain_skeleton_file(self):
        self._create_skeleton_file("domain_template.xml", "domain.xml")

    def _create_service_interface_skeleton_file(self):
        self._create_skeleton_file("service_interface_template.xml", "service_interface.xml")

    def _ddd_path(self):
        return os.path.join(self._target_dir, self._ddd_name)

    def _create_empty_file(self, filename):
        path = os.path.join(self._ddd_path(), filename)
        open(path, 'w').close()

    def _create_directory_inside_ddd(self, directory):
        os.mkdir(f"{self._target_dir}/{self._ddd_name}/{directory}")

    def _create_interaction_tests(self):
        self._create_directory_inside_ddd("test")
        template_filename = "interaction_tests_sem_template.txt"
        ddd_relative_path = "test/interaction_tests_sem.txt"
        self._create_skeleton_file(template_filename, ddd_relative_path)

    def _create_skeleton_file(self, template_filename, ddd_relative_path):
        target = os.path.join(self._ddd_path(), ddd_relative_path)
        content = self._template_from_file(template_filename)
        utils.write_template_to_file(target, content)

    def _template_from_file(self, filename):
        path = os.path.join(utils.TEMPLATES_PATH, filename)
        content = StringIO()
        with open(path) as template:
            for line in template:
                line = line.replace('__app__', self._ddd_name)
                line = line.replace('__lang__', "sem")
                line = line.replace('__App__', self._class_name_prefix)
                content.write(line)
        return content



import unittest

from unittest.mock import Mock

from tala.model.predicate import Predicate
from tala.model.sort import Sort
from tala.model.individual import Individual
from tala.model.proposition import Proposition, PredicateProposition
from tala.model.polarity import Polarity

from tala import http_formatting


class HttpFormattingTest(unittest.TestCase):
    def test_fact_to_json_object_for_value_with_grammar_entry(self):
        self.when_call_fact_to_json_object(
            self.mock_predicate_proposition(
                predicate_name="mock_predicate",
                sort_name="mock_sort",
                value="mock_individual_value",
                value_as_json_object={"value": "mock_value"}
            ), {
                "session_id": "blah",
                "entities": [{
                    "sort": "mock_sort",
                    "natural_language_form": "mock_grammar_entry",
                    "name": "mock_individual_value"
                }]
            }
        )
        self.then_result_is({
            "sort": "mock_sort",
            "value": "mock_value",
            "grammar_entry": "mock_grammar_entry",
            'perception_confidence': None,
            'understanding_confidence': None,
            'weighted_confidence': None,
            'weighted_understanding_confidence': None
        })

    def when_call_fact_to_json_object(self, fact, session):
        self._actual_result = http_formatting.fact_to_json_object(fact, session)

    def mock_predicate_proposition(
        self,
        predicate_name="mock_predicate",
        sort_name="mock_sort",
        value="mock_value",
        perception_confidence=None,
        understanding_confidence=None,
        weighted_confidence=None,
        weighted_understanding_confidence=None,
        value_as_json_object=None,
        polarity=Polarity.POS
    ):
        value_as_json_object = value_as_json_object or {"value": "mock_value"}
        sort = self._mock_sort(sort_name)
        predicate = self._mock_predicate(predicate_name, sort)
        individual = self._mock_individual(value, value_as_json_object, sort)
        mock_proposition = Mock(spec=PredicateProposition)
        mock_proposition.is_predicate_proposition.return_value = True
        mock_proposition.predicate = predicate
        mock_proposition.individual = individual
        mock_proposition.sort = sort
        mock_proposition.get_polarity.return_value = polarity
        mock_proposition.confidence_estimates.perception_confidence = perception_confidence
        mock_proposition.confidence_estimates.understanding_confidence = understanding_confidence
        mock_proposition.confidence_estimates.weighted_confidence = weighted_confidence
        mock_proposition.confidence_estimates.weighted_understanding_confidence = weighted_understanding_confidence
        return mock_proposition

    def _mock_predicate(self, name, sort):
        mock_predicate = Mock(spec=Predicate)
        mock_predicate.get_name.return_value = name
        mock_predicate.sort = sort
        return mock_predicate

    def _mock_sort(self, name):
        mock_sort = Mock(spec=Sort)
        mock_sort.get_name.return_value = name
        return mock_sort

    def _mock_individual(self, value, value_as_json_object, sort):
        mock_individual = Mock(spec=Individual)
        mock_individual.sort = sort
        mock_individual.value = value
        mock_individual.value_as_json_object.return_value = value_as_json_object
        return mock_individual

    def then_result_is(self, expected):
        self.assertEqual(expected, self._actual_result)

    def test_fact_to_json_object_for_value_without_grammar_entry(self):
        self.when_call_fact_to_json_object(
            self.mock_predicate_proposition(
                predicate_name="mock_predicate",
                sort_name="mock_sort",
                value="mock_value",
                value_as_json_object={"value": "mock_value"}
            ), {}
        )
        self.then_result_is({
            "sort": "mock_sort",
            "value": "mock_value",
            "grammar_entry": None,
            'perception_confidence': None,
            'understanding_confidence': None,
            'weighted_confidence': None,
            'weighted_understanding_confidence': None
        })

    def test_fact_to_json_object_for_none(self):
        self.when_call_fact_to_json_object(None, {})
        self.then_result_is(None)

    def test_facts_to_json_object_base_case(self):
        self.when_call_facts_to_json_object([
            self.mock_predicate_proposition(
                predicate_name="mock_predicate",
                sort_name="mock_sort",
                value="mock_value",
                value_as_json_object={"value": "mock_value"}
            )
        ], {
            "session_id": "blah",
            "entities": [{
                "sort": "mock_sort",
                "natural_language_form": "mock_grammar_entry",
                "name": "mock_value"
            }]
        })
        self.then_result_is({
            "mock_predicate": {
                "sort": "mock_sort",
                "value": "mock_value",
                "grammar_entry": "mock_grammar_entry",
                'perception_confidence': None,
                'understanding_confidence': None,
                'weighted_confidence': None,
                'weighted_understanding_confidence': None
            }
        })

    def when_call_facts_to_json_object(self, facts, session):
        self._actual_result = http_formatting.facts_to_json_object(facts, session)

    def test_facts_to_json_object_excludes_negative_predicate_propositions(self):
        self.when_call_facts_to_json_object([self.mock_predicate_proposition(polarity=Polarity.NEG)], {})
        self.then_result_is({})

    def test_facts_to_json_object_excludes_non_predicate_propositions(self):
        self.when_call_facts_to_json_object([self.mock_non_predicate_proposition()], {})
        self.then_result_is({})

    def test_facts_to_json_object_excludes_nullary_predicate_propositions(self):
        self.when_call_facts_to_json_object([self.mock_nullary_predicate_proposition()], {})
        self.then_result_is({})

    def mock_non_predicate_proposition(self, polarity=Polarity.POS):
        mock_proposition = Mock(spec=Proposition)
        mock_proposition.is_predicate_proposition.return_value = False
        mock_proposition.get_polarity.return_value = polarity
        return mock_proposition

    def mock_nullary_predicate_proposition(self, predicate_name="mock_yn_predicate", sort_name="boolean"):
        sort = self._mock_sort(sort_name)
        predicate = self._mock_predicate(predicate_name, sort)
        mock_proposition = Mock(spec=PredicateProposition)
        mock_proposition.predicate = predicate
        mock_proposition.value = None
        return mock_proposition



import re
import json


def content_matches_pattern(actual, expected):
    pattern_content = re.escape(expected).replace(r'\*', '.*')
    re_pattern = f"^{pattern_content}$"
    return re.search(re_pattern, actual, re.MULTILINE | re.DOTALL)


class StringComparison:
    def __init__(self, actual, expected):
        self._actual = actual
        self._expected = expected

    def match(self):
        self._match = self._content_matches_pattern(self._actual, self._expected)
        return self._match

    def _content_matches_pattern(self, actual, expected):
        try:
            return content_matches_pattern(actual, expected)
        except TypeError:
            pass
        for expected_alt in expected:
            match = content_matches_pattern(actual, expected_alt)
            if match:
                self._match = match
                break
        self._expected = expected_alt
        return match

    def mismatch_description(self):
        return f"""
expected: "{self._expected}"
           {self._mismatch_position_description_for_pattern()}
but got:  "{self._actual}"
           {self._mismatch_position_description_for_actual()}"""

    def _mismatch_position_description_for_pattern(self):
        position = self._get_mismatch_position()
        return " " * position + "^"

    def _get_mismatch_position(self):
        length = len(self._expected) - 1
        while length > 0:
            if self._content_matches_pattern(self._actual, self._expected[0:length] + "*"):
                return length
            length = length - 1
        return 0

    def _mismatch_position_description_for_actual(self):
        position = self._get_mismatch_position()
        return " " * position + "^"


class MoveComparison:
    def __init__(self, actual, expected):
        self._actual = actual
        self._expected = expected
        self._match = self._content_matches_pattern(actual, expected)

    def match(self):
        return self._match

    def mismatch_description(self):
        return f"""
expected: {json.dumps(self._expected)}
          {self._mismatch_position_description(self._expected, self._actual)}
but got:  {json.dumps(self._actual)}
          {self._mismatch_position_description(self._actual, self._expected)}"""

    def _content_matches_pattern(self, actual, expected):
        if len(actual) == len(expected):
            for actual_move, expected_move in zip(actual, expected):
                if not self._move_matches_expectation(actual_move, expected_move):
                    return False
            return True
        return False

    def _move_matches_expectation(self, actual_move, expected_move):
        return content_matches_pattern(actual_move, expected_move)

    def _mismatch_position_description(self, target, pattern):
        list_position = self._get_mismatch_list_position(target, pattern)
        string_position = self._get_mismatch_string_position(target, list_position)
        length = self._get_mismatch_length(target, list_position)
        if string_position > 0:
            return " " * string_position + "^" * length
        else:
            return "^" * length

    def _get_mismatch_list_position(self, one_list, other_list):
        if len(one_list) == len(other_list):
            return self._get_mismatch_list_position_for_equal_length_move_lists(one_list, other_list)
        else:
            return self._get_mismatch_list_position_for_one_move_list_longer(one_list, other_list)

    def _get_mismatch_list_position_for_equal_length_move_lists(self, one_list, other_list):
        index = 0
        for one, other in zip(one_list, other_list):
            if not one == other:
                return index
            index += 1

    def _get_mismatch_list_position_for_one_move_list_longer(self, one_list, other_list):
        shortest_list_length = min(len(one_list), len(other_list))
        for index in range(0, shortest_list_length):
            if not one_list[index] == other_list[index]:
                return index
        return shortest_list_length

    def _get_mismatch_string_position(self, target_list, list_position):
        if list_position > len(target_list):
            return len(str(target_list))
        position = 1
        for index in range(0, list_position):
            position += len(target_list[index]) + 4
        return position

    def _get_mismatch_length(self, target, position):
        if position >= len(target):
            return 1
        else:
            return len(target[position]) + 2



from tala.model import user_move


class TestUserMove:
    def test_create_ddd_specific_user_move_from_json(self):
        self.given_user_move_as_json({
            'ddd': 'some-ddd',
            'perception_confidence': 1.0,
            'semantic_expression': 'ask(?X.qna_response(X))',
            'understanding_confidence': 1.0,
        })
        self.when_create_user_move()
        self.then_move_as_dict_equals_original_goal()

    def given_user_move_as_json(self, json_dict):
        self._move_as_json_dict = json_dict

    def when_create_user_move(self):
        self._user_move = user_move.create(self._move_as_json_dict)

    def then_move_as_dict_equals_original_goal(self):
        assert self._move_as_json_dict == self._user_move.as_dict()

    def test_create_non_ddd_specific_user_move_from_json(self):
        self.given_user_move_as_json({
            'perception_confidence': 1.0,
            'semantic_expression': 'ask(?X.qna_response(X))',
            'understanding_confidence': 1.0,
        })
        self.when_create_user_move()
        self.then_move_as_dict_equals_original_goal()



import pytest

from tala.ddd.ddd_manager import DDDManager
from tala.ddd.loading.extended_ddd_set_loader import ExtendedDDDSetLoader
from tala.testing import utils as test_utils


@pytest.fixture(scope="class")
def ddd_manager(request):
    request.cls.ddd_manager = DDDManager()


@pytest.fixture(scope="class")
def loaded_mockup_travel(request):
    extended_ddd_set_loader = ExtendedDDDSetLoader(request.cls.ddd_manager)
    test_utils.load_mockup_travel(extended_ddd_set_loader)



# TODO:
# Sorts as at

end = {
    "data": {
        "type": "tala.model.domain.Domain",
        "id": "mockup_domain",
        "version:id": "2",
        "relationships": {
            "plans": [{
                "data": {
                    "type": "tala.model.plan",
                    "id": "mockup_domain:resolve:price"
                }
            }]
        },
    },
    "included": [{
        "type": "tala.model.plan",
        "id": "mockup_domain:resolve:price",
        "version:id": "2",
        "attributes": {
            "ontology_name": "mockup_ontology",
            "name": "paris"
        },
        "relationships": {
            "content": {
                "data": [{
                    "type": "tala.model.plan_item",
                    "id": "greet"
                }, {
                    "type": "tala.model.plan_item",
                    "id": "mockup_ontology:assume:some_predicate:some_individual"
                }]
            }
        },
        "included": [{
            "type": "tala.model.plan_item",
            "id": "greet",
            "version:id": "2",
            "attributes": {},
            "relationships": {}
        }, {
            "type": "tala.model.plan_item",
            "id": "mockup_ontology:assume:some_predicate:some_individual",
            "version:id": "2",
            "attributes": {},
            "relationships": {
                "content": {
                    "data": {
                        "type": "tala.model.proposition",
                        "id": "some_predicate:some_individual"
                    }
                }
            }
        }, {
            "type": "tala.model.sort",
            "id": "mockup_ontology:city",
            "attributes": {
                "version:id": "2",
                "ontology_name": "mockup_ontology",
                "name": "city",
                "dynamic": False
            }
        }]
    }]
}

input = {
    "data": [{
        "type": "tala.model.domain.Domain",
        "id": "mockup_ddd",
        "attributes": {
            "user_defined_name": "Domain Name",
            "id": "tala.model.domain.Domain.user_defined_name",
            "version:id": "2",
        },
        "relationships": [{
            "type": "tala.model.domain.Domain.default_question",
            "id": "tala.model.domain.Domain.default_question",
            "version:id": "2",
            "attributes": [],
            "relationships": {
                "default_question": {
                    "data": {
                        "type": "tala.model.question.WhQuestion",
                        "id": "mockup_ontology:WHQ:X.dest_city(X)",
                        "version:id": "2",
                        "attributes": {
                            "ontology_name": "mockup_ontology",
                            "type": "WHQ"
                        },
                        "relationships": {
                            "content": {
                                "data": {
                                    "type": "LambdaAbstractedPredicateProposition",
                                    "id": "X.dest_city(X)",
                                    "version:id": "2",
                                    "relationships": {
                                        "predicate": {
                                            "data": {
                                                "type": "tala.model.predicate",
                                                "id": "mockup_ontology:dest_city",
                                                "version:id": "2",
                                                "attributes": {
                                                    "ontology_name": "mockup_ontology",
                                                    "name": "dest_city"
                                                },
                                                "relationships": {
                                                    "sort": {
                                                        "data": {
                                                            "type": "tala.model.sort",
                                                            "id": "mockup_ontology:city",
                                                            "version:id": "2",
                                                            "attributes": {
                                                                "ontology_name": "mockup_ontology",
                                                                "name": "city",
                                                                "dynamic": True
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }]
    }]
}

output = {
    "data": [{
        "type": "tala.model.domain.Domain",
        "id": "mockup_ddd",
        "attributes": {
            "version:id": "2",
            "user_defined_name": "MockDDD"
        },
        "relationships": [{
            "type": "tala.model.domain.Domain.default_question",
            "id": "tala.model.domain.Domain.default_question",
            "version:id": "2",
            "attributes": [],
            "relationships": {
                "default_question": {
                    "data": {
                        "type": "tala.model.question.WhQuestion",
                        "id": "mockup_ontology:WHQ:X.dest_city(X)"
                    }
                }
            }
        }]
    }],
    "included": [{
        "type": "tala.model.question.WhQuestion",
        "id": "mockup_ontology:WHQ:X.dest_city(X)",
        "version:id": "2",
        "attributes": {
            "ontology_name": "mockup_ontology",
            "type": "WHQ"
        },
        "relationships": {
            "content": {
                "data": {
                    "type": "LambdaAbstractedPredicateProposition",
                    "id": "X.dest_city(X)",
                }
            }
        }
    }, {
        "type": "LambdaAbstractedPredicateProposition",
        "id": "X.dest_city(X)",
        "version:id": "2",
        "relationships": {
            "predicate": {
                "data": {
                    "type": "tala.model.predicate",
                    "id": "mockup_ontology:dest_city",
                }
            }
        }
    }, {
        "type": "tala.model.predicate",
        "id": "mockup_ontology:dest_city",
        "attributes": {
            "ontology_name": "mockup_ontology",
            "name": "dest_city"
        },
        "relationships": {
            "sort": {
                "data": {
                    "type": "tala.model.sort",
                    "id": "city"
                }
            }
        }
    }, {
        "type": "tala.model.sort",
        "id": "city",
        "attributes": {
            "ontology_name": "mockup_ontology",
            "name": "city",
            "dynamic": True
        }
    }]
}

s = {
    'data': [{
        'type': 'tala.model.ontology',
        'id': 'mockup_ontology',
        'version:id': '2',
        'attributes': {
            'ontology_name': 'mockup_ontology',
            'individuals': [],
            'actions': []
        },
        'relationships': {
            'predicates': {
                "data": [{
                    'type': 'tala.model.predicate',
                    'id': 'mockup_ontology:third predicate',
                }, {
                    'type': 'tala.model.predicate',
                    'id': 'mockup_ontology:first predicate',
                }, {
                    'type': 'tala.model.predicate',
                    'id': 'second predicate',
                }],
            }
        }
    }],
    'included': [{
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:another sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'another sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'real',
        'attributes': {
            'version:id': '2',
            'name': 'real',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.predicate',
        'id': 'mockup_ontology:third predicate',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'id': 'third predicate'
        },
        'relationships': {
            'sort': {
                'data': {
                    'type': 'tala.model.sort',
                    'id': 'mockup_ontology:another sort'
                }
            }
        }
    }, {
        'type': 'tala.model.predicate',
        'id': 'mockup_ontology:first predicate',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'id': 'first predicate'
        },
        'relationships': {
            'sort': {
                'data': {
                    'type': 'tala.model.sort',
                    'id': 'mockup_ontology:sort'
                }
            }
        }
    }, {
        'type': 'tala.model.predicate',
        'id': 'second predicate',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'id': 'second predicate'
        },
        'relationships': {
            'sort': {
                'data': {
                    'type': 'tala.model.sort',
                    'id': 'real'
                }
            }
        }
    }]
}

t = {
    'data': [{
        'type': 'tala.model.ontology',
        'id': 'mockup_ontology',
        'version:id': '2',
        'attributes': {
            'ontology_name': 'mockup_ontology',
            'id': 'mockup_ontology',
            'predicates': [{
                'type': 'tala.model.predicate',
                'id': 'mockup_ontology:another predicate 1',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'another predicate 1'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:another sort'
                        }
                    }
                }
            }, {
                'type': 'tala.model.predicate',
                'id': 'mockup_ontology:a predicate 2',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'a predicate 2'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:a sort'
                        }
                    }
                }
            }, {
                'type': 'tala.model.predicate',
                'id': 'mockup_ontology:another predicate 2',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'another predicate 2'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:another sort'
                        }
                    }
                }
            }, {
                'type': 'tala.model.predicate',
                'id': 'mockup_ontology:a predicate 1',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'a predicate 1'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:a sort'
                        }
                    }
                }
            }],
            'individuals': [{
                'type': 'tala.model.individual',
                'id': 'mockup_ontology:individual',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'individual'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:a sort'
                        }
                    }
                }
            }],
            'actions': []
        },
        'relationships': {}
    }],
    'included': [{
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:another sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'another sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:a sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'a sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:another sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'another sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:a sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'a sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:a sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'a sort',
            'dynamic': False
        }
    }]
}

u = {
    'data': [{
        'type': 'tala.model.ontology',
        'id': 'mockup_ontology',
        'version:id': '2',
        'attributes': {
            'ontology_name': 'mockup_ontology',
            'name': 'mockup_ontology'
        },
        'relationships': {
            'predicates': [{
                'type': 'tala.model.predicate',
                'id': 'mockup_ontology:another predicate 2',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'another predicate 2'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:another sort'
                        }
                    }
                }
            }, {
                'type': 'tala.model.predicate',
                'id': 'mockup_ontology:a predicate 1',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'a predicate 1'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:a sort'
                        }
                    }
                }
            }, {
                'type': 'tala.model.predicate',
                'id': 'mockup_ontology:another predicate 1',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'another predicate 1'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:another sort'
                        }
                    }
                }
            }, {
                'type': 'tala.model.predicate',
                'id': 'mockup_ontology:a predicate 2',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'a predicate 2'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:a sort'
                        }
                    }
                }
            }],
            'sorts': [{
                'type': 'tala.model.sort',
                'id': 'mockup_ontology:another sort',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'another sort',
                    'dynamic': False
                }
            }, {
                'type': 'tala.model.sort',
                'id': 'mockup_ontology:a sort',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'a sort',
                    'dynamic': False
                }
            }],
            'individuals': [{
                'type': 'tala.model.individual',
                'id': 'mockup_ontology:individual',
                'attributes': {
                    'version:id': '2',
                    'ontology_name': 'mockup_ontology',
                    'name': 'individual'
                },
                'relationships': {
                    'sort': {
                        'data': {
                            'type': 'tala.model.sort',
                            'id': 'mockup_ontology:a sort'
                        }
                    }
                }
            }],
            'actions': []
        }
    }],
    'included': [{
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:another sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'another sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:a sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'a sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:another sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'another sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:a sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'a sort',
            'dynamic': False
        }
    }, {
        'type': 'tala.model.sort',
        'id': 'mockup_ontology:a sort',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'a sort',
            'dynamic': False
        }
    }]
}

p = {
    'data': {
        'type': 'tala.model.ontology',
        'id': 'mockup_ontology',
        'version:id': '2',
        'attributes': {},
        'relationships': {
            'predicates': {
                'data': [{
                    'type': 'tala.model.predicate',
                    'id': 'mockup_ontology:a predicate 1'
                }, {
                    'type': 'tala.model.predicate',
                    'id': 'mockup_ontology:a predicate 2'
                }, {
                    'type': 'tala.model.predicate',
                    'id': 'mockup_ontology:another predicate 1'
                }, {
                    'type': 'tala.model.predicate',
                    'id': 'mockup_ontology:another predicate 2'
                }]
            },
            'individuals': {
                'data': [{
                    'type': 'tala.model.individual',
                    'id': 'mockup_ontology:individual'
                }]
            },
            'sorts': {
                'data': []
            },
            'actions': {
                'data': []
            }
        }
    },
    'included': [{
        'type': 'tala.model.predicate',
        'id': 'mockup_ontology:a predicate 1',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'a predicate 1'
        },
        'relationships': {
            'sort': {
                'data': {
                    'type': 'tala.model.sort',
                    'id': 'mockup_ontology:a sort'
                }
            }
        }
    }, {
        'type': 'tala.model.predicate',
        'id': 'mockup_ontology:a predicate 2',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'a predicate 2'
        },
        'relationships': {
            'sort': {
                'data': {
                    'type': 'tala.model.sort',
                    'id': 'mockup_ontology:a sort'
                }
            }
        }
    }, {
        'type': 'tala.model.predicate',
        'id': 'mockup_ontology:another predicate 1',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'another predicate 1'
        },
        'relationships': {
            'sort': {
                'data': {
                    'type': 'tala.model.sort',
                    'id': 'mockup_ontology:another sort'
                }
            }
        }
    }, {
        'type': 'tala.model.predicate',
        'id': 'mockup_ontology:another predicate 2',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'another predicate 2'
        },
        'relationships': {
            'sort': {
                'data': {
                    'type': 'tala.model.sort',
                    'id': 'mockup_ontology:another sort'
                }
            }
        }
    }, {
        'type': 'tala.model.individual',
        'id': 'mockup_ontology:individual',
        'attributes': {
            'version:id': '2',
            'ontology_name': 'mockup_ontology',
            'name': 'individual'
        },
        'relationships': {
            'sort': {
                'data': {
                    'type': 'tala.model.sort',
                    'id': 'mockup_ontology:a sort'
                }
            }
        }
    }]
}



from tala.utils.equality import EqualityMixin


class PersonName(EqualityMixin):
    def __init__(self, string):
        self._string = string

    @property
    def string(self):
        return self._string

    def __str__(self):
        return "person_name(%s)" % self._string

    def __repr__(self):
        return 'PersonName(%r)' % self._string



from typing import Text  # noqa: F401

from tala.utils.equality import EqualityMixin


def create(user_move_as_dict):
    try:
        return DDDSpecificUserMove.from_dict(user_move_as_dict)
    except KeyError:
        return UserMove.from_dict(user_move_as_dict)


class UserMove(EqualityMixin):
    @classmethod
    def from_dict(cls, move_as_json):
        perception_confidence = move_as_json["perception_confidence"]
        understanding_confidence = move_as_json["understanding_confidence"]
        semantic_expression = move_as_json["semantic_expression"]

        return cls(semantic_expression, perception_confidence, understanding_confidence)

    def __init__(self, semantic_expression, perception_confidence, understanding_confidence):
        # type: (Text, float, float) -> None
        self._semantic_expression = semantic_expression
        self._perception_confidence = perception_confidence
        self._understanding_confidence = understanding_confidence

    @property
    def is_ddd_specific(self):
        # type: () -> bool
        return False

    @property
    def semantic_expression(self):
        # type: () -> Text
        return self._semantic_expression

    @property
    def perception_confidence(self):
        # type: () -> float
        return self._perception_confidence

    @property
    def understanding_confidence(self):
        # type: () -> float
        return self._understanding_confidence

    def as_dict(self):
        return {
            "perception_confidence": self.perception_confidence,
            "understanding_confidence": self.understanding_confidence,
            "semantic_expression": self.semantic_expression,
        }

    def __str__(self):
        return f"{self.__class__.__name__}({self._semantic_expression}, " \
               f"perception_confidence={self._perception_confidence}, " \
               f"understanding_confidence={self._understanding_confidence})"

    def __repr__(self):
        return str(self)


class DDDSpecificUserMove(UserMove):
    @classmethod
    def from_dict(cls, move_as_json):
        ddd = move_as_json["ddd"]
        perception_confidence = move_as_json["perception_confidence"]
        understanding_confidence = move_as_json["understanding_confidence"]
        semantic_expression = move_as_json["semantic_expression"]

        return cls(ddd, semantic_expression, perception_confidence, understanding_confidence)

    def __init__(self, ddd, semantic_expression, perception_confidence, understanding_confidence):
        # type: (Text, Text, float, float) -> None
        super(DDDSpecificUserMove, self).__init__(semantic_expression, perception_confidence, understanding_confidence)
        self._ddd = ddd

    @property
    def is_ddd_specific(self):
        # type: () -> bool
        return True

    @property
    def ddd(self):
        # type: () -> Text
        return self._ddd

    def as_dict(self):
        return {
            "ddd": self.ddd,
            "perception_confidence": self.perception_confidence,
            "understanding_confidence": self.understanding_confidence,
            "semantic_expression": self.semantic_expression,
        }

    def __str__(self):
        return f"{self.__class__.__name__}({self._ddd}, semantic_expression={self._semantic_expression}, " \
               f"perception_confidence={self._perception_confidence}, " \
               f"understanding_confidence={self._understanding_confidence})"



FRONTEND_CONNECTION_PATH = r"/interact"
INTERMITTENT_WAIT_TIMEOUT = 5  # seconds
DDD_DB_NAME = "ddds"
PROTOCOL_VERSION = "3.4"
DEPRECATED_PROTOCOL_VERSIONS = [2.0, "3.0", "3.1", "3.2", "3.3"]
SUPPORTED_PROTOCOL_VERSIONS = DEPRECATED_PROTOCOL_VERSIONS + [PROTOCOL_VERSION]

DEFAULT_PORT = 9090



class EqualityMixin(object):
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        x = self.__eq__(other)
        if x is not NotImplemented:
            return not x
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(items(self))))


def items(object_):
    if isinstance(object_, (list, tuple, set)):
        return (items(element) for element in object_)
    if isinstance(object_, dict):
        return tuple({key: items(value) for key, value in list(object_.items())})
    if hasattr(object_, "__class__"):
        if not isinstance(object_, EqualityMixin):
            if hasattr(object_, "__hash__"):
                return hash(object_)
    if hasattr(object_, "__dict__"):
        return items(object_.__dict__)
    return object_



from tala.model.semantic_object import OntologySpecificSemanticObject
from tala.model.sort import Sort
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin


class Predicate(OntologySpecificSemanticObject, AsSemanticExpressionMixin):
    @classmethod
    def create_from_json_api_data(cls, predicate_entry, included):
        sort_entry = included.get_object_from_relationship(predicate_entry["relationships"]["sort"]["data"])
        sort = Sort.create_from_json_api_data(sort_entry, included)
        ontology_name = predicate_entry["attributes"]["ontology_name"]
        name = predicate_entry["attributes"]["name"]
        feature_of_name = predicate_entry["attributes"]["feature_of_name"]
        multiple_instances = predicate_entry["attributes"]["_multiple_instances"]
        return cls(ontology_name, name, sort, feature_of_name, multiple_instances)

    def __init__(self, ontology_name, name, sort, feature_of_name=None, multiple_instances=False):
        OntologySpecificSemanticObject.__init__(self, ontology_name)
        self.name = name
        self.sort = sort
        self.feature_of_name = feature_of_name
        self._multiple_instances = multiple_instances

    def get_name(self):
        return self.name

    def get_feature_of_name(self):
        return self.feature_of_name

    def is_feature_of(self, other):
        return self.feature_of_name == other.get_name()

    def allows_multiple_instances(self):
        return self._multiple_instances

    def is_predicate(self):
        return True

    @property
    def json_api_id(self):
        return f"{self.ontology_name}:{self.name}"

    @property
    def json_api_attributes(self):
        return ["ontology_name", "name", "feature_of_name", "_multiple_instances"]

    @property
    def json_api_relationships(self):
        return ["sort"]

    def __eq__(self, other):
        try:
            return other.get_name() == self.get_name() \
                and other.sort == self.sort \
                and other.get_feature_of_name() == self.get_feature_of_name() \
                and other.allows_multiple_instances() == self.allows_multiple_instances()
        except AttributeError:
            return False

    def __lt__(self, other):
        return self.name < other.name

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.name, self.sort, self.feature_of_name, self._multiple_instances))

    def __str__(self):
        return self.get_name()

    def __repr__(self):
        return "%s%s" % (
            self.__class__.__name__, (self.name, self.sort, self.feature_of_name, self._multiple_instances)
        )



from tala.utils import json_api

uuid_counter = 0


def reset_mock_uuid():
    global uuid_counter
    uuid_counter = 0


def mock_uuid():
    global uuid_counter
    uuid_counter += 1
    return f"uuid_{uuid_counter}"


class TestJSONAPI:
    def setup_method(self):
        self._original_uuid = json_api.uuid.uuid4
        json_api.uuid.uuid4 = mock_uuid
        reset_mock_uuid()

    def teardown_method(self):
        json_api.uuid.uuid4 = self._original_uuid

    def test_create_object(self):
        self.when_create_json_api("some-type")
        self.then_dict_is({
            "data": {
                "type": "some-type",
                "id": "uuid_1",
                "attributes": {},
                "relationships": {},
                "version:id": "2"
            },
            "included": []
        })

    def when_create_json_api(self, type_, id_=None):
        self._json_api = json_api.JSONAPIObject(type_, id_)

    def then_dict_is(self, data):
        assert data == self._json_api.as_dict

    def test_create_object_with_id(self):
        self.when_create_json_api("some-type", "some-id")
        self.then_dict_is({
            "data": {
                "type": "some-type",
                "id": "some-id",
                "attributes": {},
                "relationships": {},
                "version:id": "2"
            },
            "included": []
        })

    def test_add_attribute(self):
        self.given_json_api_created()
        self.when_add_attribute("name", "value")
        self.then_dict_is({
            "data": {
                "type": "some-type",
                "id": "uuid_1",
                "attributes": {
                    "name": "value"
                },
                "relationships": {},
                "version:id": "2"
            },
            "included": []
        })

    def given_json_api_created(self, type_="some-type"):
        self._json_api = json_api.JSONAPIObject(type_)

    def when_add_attribute(self, name, value):
        self._json_api.add_attribute(name, value)

    def test_add_relationship(self):
        self.given_json_api_created("root-type")
        self.when_add_relationship(
            "included-data", {
                "data": {
                    "type": "resource-type",
                    "id": "some-id",
                    "attributes": {
                        "name": "value"
                    },
                    "relationships": {},
                    "version:id": "2"
                },
                "included": []
            }
        )
        self.then_dict_is({
            'data': {
                'attributes': {},
                'id': 'uuid_1',
                'relationships': {
                    'included-data': {
                        'data': {
                            'id': 'some-id',
                            'type': 'resource-type',
                        },
                    },
                },
                'type': 'root-type',
                'version:id': '2',
            },
            'included': [
                {
                    'attributes': {
                        'name': 'value',
                    },
                    'id': 'some-id',
                    'relationships': {},
                    'type': 'resource-type',
                    'version:id': '2',
                },
            ],
        })

    def when_add_relationship(self, name, data):
        self._json_api.add_relationship(name, data)

    def test_add_attribute_list(self):
        self.given_json_api_created()
        self.when_add_attribute("list-attribute", [])
        self.then_dict_is({
            'data': {
                'attributes': {
                    'list-attribute': []
                },
                'id': 'uuid_1',
                'relationships': {},
                'type': 'some-type',
                'version:id': '2'
            },
            'included': []
        })

    def test_append_attribute(self):
        self.given_json_api_created()
        self.when_append_attribute("list-attribute", "item")
        self.then_dict_is({
            'data': {
                'attributes': {
                    'list-attribute': ['item']
                },
                'id': 'uuid_1',
                'relationships': {},
                'type': 'some-type',
                'version:id': '2'
            },
            'included': []
        })

    def given_attribute_added(self, name, value):
        self._json_api.add_attribute(name, value)

    def when_append_attribute(self, name, value):
        self._json_api.append_attribute(name, value)

    def test_append_relationship(self):
        self.given_json_api_created()
        self.when_append_relationship(
            "list-relationship", {
                'data': {
                    'attributes': {
                        'list-attribute': ['item']
                    },
                    'id': 'uuid_1',
                    'relationships': {},
                    'type': 'some-type',
                    'version:id': '2'
                },
                'included': []
            }
        )
        self.then_dict_is({
            'data': {
                'attributes': {},
                'id': 'uuid_1',
                'relationships': {
                    'list-relationship': {
                        'data': [{
                            'id': 'uuid_1',
                            'type': 'some-type'
                        }]
                    }
                },
                'type': 'some-type',
                'version:id': '2'
            },
            'included': [{
                'attributes': {
                    'list-attribute': ['item']
                },
                'id': 'uuid_1',
                'relationships': {},
                'type': 'some-type',
                'version:id': '2'
            }]
        })

    def when_append_relationship(self, name, data):
        self._json_api.append_relationship(name, data)

    def test_add_resource_object_with_relationships(self):
        self.given_json_api_created("root-type")
        self.when_add_relationship(
            "root-included", {
                'data': {
                    'attributes': {},
                    'id': 'uuid_1',
                    'relationships': {
                        'included-data': {
                            'data': {
                                'id': 'some-id',
                                'type': 'resource-type',
                            },
                        },
                    },
                    'type': 'root-type',
                    'version:id': '2',
                },
                'included': [
                    {
                        'attributes': {
                            'name': 'value',
                        },
                        'id': 'some-id',
                        'relationships': {},
                        'type': 'resource-type',
                        'version:id': '2',
                    },
                ],
            }
        )
        self.then_dict_is({
            'data': {
                'attributes': {},
                'id': 'uuid_1',
                'relationships': {
                    'root-included': {
                        'data': {
                            'id': 'uuid_1',
                            'type': 'root-type'
                        }
                    }
                },
                'type': 'root-type',
                'version:id': '2'
            },
            'included': [{
                'attributes': {
                    'name': 'value'
                },
                'id': 'some-id',
                'relationships': {},
                'type': 'resource-type',
                'version:id': '2'
            }, {
                'attributes': {},
                'id': 'uuid_1',
                'relationships': {
                    'included-data': {
                        'data': {
                            'id': 'some-id',
                            'type': 'resource-type'
                        }
                    }
                },
                'type': 'root-type',
                'version:id': '2'
            }]
        })



import unittest

from tala.model import speaker


class speakerTests(unittest.TestCase):
    def test_speaker_class(self):
        self.assertEqual("SYS", speaker.SYS)
        self.assertEqual("USR", speaker.USR)



import logging.config
import re
import os
import copy

from structlog.processors import JSONRenderer, TimeStamper, StackInfoRenderer, format_exc_info
from structlog.stdlib import add_log_level, filter_by_level, add_logger_name
import structlog


class EnvironmentVariableNotDefinedException(Exception):
    pass


def getenv(key, default=None):
    value = os.getenv(key, default)
    if value:
        return value
    if default is None:
        raise EnvironmentVariableNotDefinedException(
            f"Expected environment variable '{key}' to be defined but it wasn't"
        )


class FuncBase:
    def _setup_logger(self, request, logger):
        def get_session_id():
            try:
                return request["session"]["session_id"]
            except KeyError:
                return "unknown-session-id"

        def get_device_id():
            try:
                return request["session"]["device_id"]
            except KeyError:
                return "unknown-device-id"

        self.logger = logger.bind(session_id=get_session_id(), device_id=get_device_id())


REDACTED_LOG_FIELDS_IN_SESSION = [
    "asr_hints", "dme_http_services", "expected_input", "nlg", "nlu", "odb", "tis", "visual_output"
]


def log_models(original_body, logger, models=REDACTED_LOG_FIELDS_IN_SESSION):
    log_body = copy.deepcopy(original_body)
    if "tis" in log_body.get("session", {}):
        del log_body["session"]["tis"]
    for field in models:
        if field in log_body.get("session", {}):
            logger.info(f"Model log for {field}", body=log_body["session"][field])


def redact_bulky_entries_in_session(_, __, input_event_dict):
    if "session" in input_event_dict.get("body", {}):
        event_dict = copy.deepcopy(input_event_dict)
        for field in REDACTED_LOG_FIELDS_IN_SESSION:
            if field in event_dict.get("body", {}).get("session", {}):
                event_dict["body"]["session"][field] = "[REDACTED]"
        return event_dict
    return input_event_dict


def configure_stdout_logging(level=None):
    structlog.configure(
        processors=[
            redact_bulky_entries_in_session,
            filter_by_level,
            add_log_level,
            add_logger_name,
            TimeStamper(fmt="iso", key="time"),
            StackInfoRenderer(),
            format_exc_info,
            JSONRenderer(sort_keys=True),
        ],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )
    logging.config.dictConfig({
        "version": 1,
        "formatters": {
            "plain": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.dev.ConsoleRenderer(colors=False),
            },
            "access-json": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.processors.JSONRenderer(),
                "foreign_pre_chain": [
                    structlog.stdlib.add_log_level,
                    structlog.stdlib.add_logger_name,
                    TimeStamper(fmt="iso", key="time"),
                    StackInfoRenderer(),
                    format_exc_info,
                    combined_logformat,
                ],
            }
        },
        "handlers": {
            "stdout": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
                "formatter": "plain",
            },
            "access": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
                "formatter": "access-json",
            },
        },
        "loggers": {
            "": {
                "handlers": ["stdout"],
                "level": level,
                "propagate": False,
            },
            "gunicorn.access": {
                "handlers": ["access"],
                "level": level,
                "propagate": False,
            },
        }
    })  # yapf: disable


def combined_logformat(logger, name, event_dict):
    if event_dict.get('logger') == "gunicorn.access":
        message = event_dict['event']

        parts = [
            r'(?P<host>\S+)',  # host %h
            r'\S+',  # indent %l (unused)
            r'(?P<user>\S+)',  # user %u
            r'\[(?P<time>.+)\]',  # time %t
            r'"(?P<request>.+)"',  # request "%r"
            r'(?P<status>[0-9]+)',  # status %>s
            r'(?P<size>\S+)',  # size %b (careful, can be '-')
            r'"(?P<referer>.*)"',  # referer "%{Referer}i"
            r'"(?P<agent>.*)"',  # user agent "%{User-agent}i"
        ]
        pattern = re.compile(r'\s+'.join(parts) + r'\s*\Z')
        m = pattern.match(message)
        result = m.groupdict()

        if result["user"] == "-":
            result["user"] = None

        result["status"] = int(result["status"])

        if result["size"] == "-":
            result["size"] = 0
        else:
            result["size"] = int(result["size"])

        if result["referer"] == "-":
            result["referer"] = None

        event_dict.update(result)
        event_dict["event"] = "endpoint accessed"

    return event_dict



import json
import threading
import queue
import uuid
import websocket

import requests

STREAMING_DONE = "STREAMING_DONE"
STREAMING_SET_PERSONA = "STREAMING_SET_PERSONA"
STREAMING_SET_VOICE = "STREAMING_SET_VOICE"
STREAMING_CHUNK = "STREAMING_CHUNK"
TIMEOUT = 5.0


class SSEClientException(BaseException):
    pass


class AbstractSSEClient:
    def __init__(self, client_id_base, logger, endpoint, port=None):
        self._setup_logger(logger)
        self._endpoint = endpoint
        if port:
            self._port = int(port)
        else:
            self._port = 443
        self._client_id = f"{client_id_base}-{str(uuid.uuid4())}"

    def _setup_logger(self, logger):
        self._base_logger = logger
        self._set_logger(logger)

    def _set_logger(self, logger):
        self.logger = logger

    def _reset_logger(self):
        self._set_logger(self._base_logger)

    def start(self):
        pass

    @property
    def client_id(self):
        return self._client_id

    def _get_topic(self, streamer_session):
        return f'tm/{streamer_session["session_id"]}'

    def open_session(self, session_id, s_and_r_dict=None, logger=None):
        streamer_thread, chunk_joiner = self._prepare_streamer_thread(session_id)

        streamer_session = {
            "session_id": session_id,
            "search_and_replace_dict": s_and_r_dict if s_and_r_dict else {},
            "message_counter": 0,
            "streamed": [],
            "logger": logger,
            "streamer_thread": streamer_thread,
            "chunk_joiner": chunk_joiner
        }
        self.sessions[session_id] = streamer_session
        streamer_session["streamer_thread"].start()

    def _apply_dictionary(self, session_id, chunk):
        def key_matches_beginning_of_chunk(key, chunk):
            return chunk.startswith(f"{key} ")

        def key_matches_middle_of_chunk(key, chunk):
            chunk_ends = [" ", ".", ",", "!", "?"]
            for end in chunk_ends:
                search_term = f" {key}{end}"
                if search_term in chunk:
                    return search_term

        def key_matches_end_of_chunk(key, chunk):
            return chunk.endswith(f" {key}")

        search_and_replace_dict = self.sessions[session_id]["search_and_replace_dict"]

        if chunk in search_and_replace_dict:
            self.logger.info("Matched entire chunk", chunk=chunk)
            return search_and_replace_dict[chunk]
        for key in search_and_replace_dict:
            if key_matches_beginning_of_chunk(key, chunk):
                self.logger.info("Matched beginning of chunk", key=key, chunk=chunk)
                chunk = chunk.replace(key, search_and_replace_dict[key], 1)
            match = key_matches_middle_of_chunk(key, chunk)
            while match:
                self.logger.info("Matched middle of chunk", key=key, chunk=chunk)
                chunk = chunk.replace(key, search_and_replace_dict[key], 1)
                match = key_matches_middle_of_chunk(key, chunk)
            if key_matches_end_of_chunk(key, chunk):
                self.logger.info("Matched end of chunk", key=key, chunk=chunk)
                chunk = chunk.replace(key, search_and_replace_dict[key], 1)
        return chunk

    def _prepare_streamer_thread(self, session_id):
        def stream_chunks():
            self._streaming_exception = None
            streamer_session = self.sessions[session_id]
            for chunk in streamer_session["chunk_joiner"]:
                try:
                    chunk = self._apply_dictionary(session_id, chunk)
                    self._stream_to_frontend(streamer_session, {"event": STREAMING_CHUNK, "data": chunk})
                    streamer_session["streamed"].append(chunk)
                except Exception as e:
                    streamer_session["streaming_exception"] = e
                    self.logger.exception("Exception raised in streamer thread")

        chunk_joiner = ChunkJoiner(self.logger)
        streamer_thread = threading.Thread(target=stream_chunks)
        return streamer_thread, chunk_joiner

    def stream_utterance(self, session_id, persona=None, voice=None, utterance=""):
        self.set_persona(session_id, persona)
        self.set_voice(session_id, voice)
        self.stream_chunk(session_id, utterance + " ")

    def set_persona(self, session_id, persona):
        streamer_session = self.sessions[session_id]
        self._stream_to_frontend(streamer_session, {"event": STREAMING_SET_PERSONA, "data": persona if persona else ""})

    def set_voice(self, session_id, voice):
        streamer_session = self.sessions[session_id]
        self._stream_to_frontend(streamer_session, {"event": STREAMING_SET_VOICE, "data": voice if voice else ""})

    def stream_chunk(self, session_id, chunk):
        streamer_session = self.sessions[session_id]
        joiner = streamer_session["chunk_joiner"]
        joiner.add_chunk(chunk)

    def flush_stream(self, session_id):
        self.logger.info("flushing stream")
        streamer_session = self.sessions[session_id]
        joiner = streamer_session["chunk_joiner"]
        joiner.last_chunk_sent()
        streamer_thread = streamer_session["streamer_thread"]
        streamer_thread.join(3.0)
        if streamer_thread.is_alive():
            self.logger.warn("Streamer thread is still alive!", streamed=self._streamed)

    def end_stream(self, session_id):
        self.logger.info("ending stream")
        streamer_session = self.sessions[session_id]
        self._stream_to_frontend(streamer_session, {"event": STREAMING_DONE})


class StreamIterator:
    def __init__(self, session_id, base_url, logger):
        self.logger = logger
        self.url = f'{base_url}/{session_id}'
        self.response = None
        self.iterator = None

        self._current_event = None
        self._session_id = session_id
        self._stream = None

    def wait_for_message(self, expected_message):
        next_message = self.get_next_message()
        if json.loads(expected_message) != json.loads(next_message):
            self.logger.info("received unexpected message", expected=expected_message, received=next_message)
            raise SSEClientException(f"Expected '{expected_message}', but received '{next_message}'.")
        else:
            self.logger.info("received expected message", message=next_message)

    @property
    def session_id(self):
        return self._session_id

    def start_stream(self):
        """Start the streaming connection and prepare the iterator."""
        self.response = requests.get(self.url, stream=True)
        self.response.raise_for_status()  # Ensure the response is successful
        self.iterator = self.response.iter_lines()

    def get_next_message(self):
        """Retrieve the next message from the stream."""
        if self.iterator is None:
            raise RuntimeError("Stream has not been started. Call start_stream() first.")
        try:
            item = self._get_next_item()
            return item
        except StopIteration:
            return None

    def close_stream(self):
        if self.response:
            self.response.close()
            self.response = None

    def _get_next_item(self):
        def is_event(line):
            return line.startswith("event: ")

        def is_data(line):
            return line.startswith("data: ")

        def extract_event(line):
            return line[len("event: "):]

        def extract_data(line):
            return line[len("data: "):]

        while True:
            line = next(self.iterator).decode('utf-8')
            if is_event(line):
                event = extract_event(line)
                if event == STREAMING_DONE:
                    self.close_stream()
                    return f'{{"event": "{event}"}}'
                elif event in [STREAMING_CHUNK, STREAMING_SET_PERSONA, STREAMING_SET_VOICE]:
                    while True:
                        next_line = next(self.iterator).decode('utf-8')
                        if is_data(next_line):
                            data = extract_data(next_line)
                            return f'{{"event": "{event}", "data": "{data.strip()}"}}'


class ChunkJoiner:
    """
    Iterates over segments put together from chunks received in the queue. A segment is not made available until we know
    that it's a complete segment. To be considered complete segment means that the next chunk starts with a space, or
    that someone called last_chunk_sent().
    """
    def __init__(self, logger):
        self.logger = logger
        self._chunk_queue = queue.Queue()
        self._received_last_chunk = threading.Event()
        self._next_chunk = None
        self._collected_chunks = []

    def add_chunk(self, chunk):
        self._chunk_queue.put(chunk)

    def last_chunk_sent(self):
        self._received_last_chunk.set()

    def __next__(self):
        return self._get_next_segment()

    def _get_next_segment(self):
        self._collected_chunks = []
        self._collect_chunks()
        if self._collected_chunks:
            next_segment = "".join(self._collected_chunks)
            self.logger.info("Joiner collected segment", next_segment=next_segment)
            return next_segment
        else:
            self.logger.info("No more chunks collected, stopping iteration")
            raise StopIteration

    def _collect_chunks(self):
        def waiting_for_first_chunk():
            return self._collected_chunks == []

        def add_next_chunk_to_collected_chunks():
            self._collected_chunks.append(self._next_chunk)
            self._next_chunk = None

        def next_chunk_belongs_to_this_segment():
            return not self._next_chunk.startswith(" ")

        def handle_possible_last_chunk():
            try:
                self._next_chunk = self._chunk_queue.get_nowait()
                if self._next_chunk:
                    self.logger.info("there was a last chunk to handle.")
                    add_next_chunk_to_collected_chunks()
            except queue.Empty:
                pass

        while not self._complete_segment_collected():
            try:
                if not self._next_chunk:
                    self._next_chunk = self._chunk_queue.get(timeout=0.05)

                if waiting_for_first_chunk() or next_chunk_belongs_to_this_segment():
                    add_next_chunk_to_collected_chunks()

            except queue.Empty:
                if self._received_last_chunk.is_set():
                    self.logger.info("received last chunk")
                    handle_possible_last_chunk()
                    break

    def _complete_segment_collected(self):
        def next_chunk_was_not_needed_to_complete_this_segment():
            return self._collected_chunks and self._next_chunk

        def collected_chunks_end_with_space():
            return self._collected_chunks and self._collected_chunks[-1].endswith(" ")

        return next_chunk_was_not_needed_to_complete_this_segment() or collected_chunks_end_with_space()

    def __iter__(self):
        return self


class SSEClient(AbstractSSEClient):
    sessions = {}

    def open_session(self, session_id, s_and_r_dict=None, logger=None):
        super().open_session(session_id, s_and_r_dict, logger)
        streamer_session = self.sessions[session_id]
        url = f"{self._endpoint}/{self._get_topic(streamer_session)}"
        session_websocket = websocket.WebSocket()
        session_websocket.connect(url)
        streamer_session["websocket"] = session_websocket

    def _stream_to_frontend(self, streamer_session, message):
        def remove_final_space(message):
            if "data" in message:
                message["data"] = message["data"].strip()
            return message

        streamer_session["message_counter"] += 1
        self.logger.debug(
            "SSE Client streaming to frontend",
            message=message,
            session_id=streamer_session["session_id"],
            client_id=self.client_id
        )
        streamer_session["websocket"].send(json.dumps(message))

    def close_session(self, session_id):
        streamer_session = self.sessions[session_id]
        streamer_session["websocket"].close()
        streamer_exception = streamer_session.get("streaming_exception")
        if streamer_exception:
            raise SSEClientException(
                f'{streamer_exception} was raised during streaming in {streamer_session["session_id"]}. Streamed: {self._streamed}.'
            )
        self.logger.info("close session", client_id=self.client_id, session_id=streamer_session["session_id"])
        self.logger.info(
            "Streamed in session",
            num_messages=streamer_session["message_counter"],
            streamed=streamer_session["streamed"]
        )
        self._reset_logger()



import pytest

from tala.utils import compression

UNCOMPRESSED_STRING = "kalle"
COMPRESSED_STRING = 'eNrLTszJSQUABiECCg=='

UNCOMPRESSED_JSON = {"name": "kalle"}
COMPRESSED_JSON = "eNqrVspLzE1VslJQyk7MyUlVqgUAMToFhQ=="


class TestStringCompression:
    def test_compress_string_base_case(self):
        self.given_string(UNCOMPRESSED_STRING)
        self.when_compress_string()
        self.then_compressed_string_is(COMPRESSED_STRING)

    def given_string(self, string):
        self._string = string

    def when_compress_string(self):
        self._compressed_string = compression.compress_string(self._string)

    def then_compressed_string_is(self, string):
        assert self._compressed_string == COMPRESSED_STRING

    def test_decompress_compressed_string_base_case(self):
        self.given_compressed_string(COMPRESSED_STRING)
        self.when_decompress_string()
        self.then_decompressed_string_is(UNCOMPRESSED_STRING)

    def given_compressed_string(self, string):
        self._compressed_string = string

    def when_decompress_string(self):
        self._decompressed_string = compression.decompress_string(self._compressed_string)

    def then_decompressed_string_is(self, string):
        assert self._decompressed_string == UNCOMPRESSED_STRING

    @pytest.mark.parametrize("data", [UNCOMPRESSED_STRING, COMPRESSED_STRING])
    def test_ensure_decompress(self, data):
        self.given_data(data)
        self.when_ensure_decompress_string()
        self.then_decompressed_string_is(UNCOMPRESSED_STRING)

    def given_data(self, string):
        self._possibly_compressed_string = string

    def when_ensure_decompress_string(self):
        self._decompressed_string = compression.ensure_decompressed_string(self._possibly_compressed_string)


class TestJSONCompression:
    def test_compress_json_dict_base_case(self):
        self.given_json(UNCOMPRESSED_JSON)
        self.when_compress_json()
        self.then_compressed_json_is(COMPRESSED_JSON)

    def given_json(self, json_dict):
        self._json = json_dict

    def when_compress_json(self):
        self._compressed_json = compression.compress_json(self._json)

    def then_compressed_json_is(self, json_dict):
        assert self._compressed_json == json_dict

    def test_decompress_compressed_json_base_case(self):
        self.given_compressed_json(COMPRESSED_JSON)
        self.when_decompress_json()
        self.then_decompressed_json_is(UNCOMPRESSED_JSON)

    def given_compressed_json(self, json_dict):
        self._compressed_json = json_dict

    def when_decompress_json(self):
        self._decompressed_json = compression.decompress_json(self._compressed_json)

    def then_decompressed_json_is(self, json_dict):
        assert self._decompressed_json == json_dict

    @pytest.mark.parametrize("data", [UNCOMPRESSED_JSON, COMPRESSED_JSON])
    def test_ensure_decompress(self, data):
        self.given_data(data)
        self.when_ensure_decompress_json()
        self.then_decompressed_json_is(UNCOMPRESSED_JSON)

    def given_data(self, json_dict):
        self._possibly_compressed_json = json_dict

    def when_ensure_decompress_json(self):
        self._decompressed_json = compression.ensure_decompressed_json(self._possibly_compressed_json)



import json
import os

from jinja2 import Environment
import structlog

from tala.utils import nlg
from tala.utils.func import configure_stdout_logging

environment = Environment()

DEFAULT_CONTEXT = {"facts": {}, "facts_being_grounded": {}, "entities_under_discussion": {}}


class TestGenerator():
    def setup_method(self):
        self.logger = structlog.get_logger(__name__)
        configure_stdout_logging("DEBUG")
        self.load_default_model()
        self.given_context(DEFAULT_CONTEXT)

    def load_default_model(self):
        self._path_of_this_script = os.path.dirname(os.path.realpath(__file__))
        self._load_model(f"{self._path_of_this_script}/nlg_data.json")

    def _load_model(self, path):
        with open(path) as nlg_data:
            self._data = json.load(nlg_data)

    def given_context(self, data):
        self._context = data

    def given_facts(self, facts):
        self._context["facts"] = facts

    def given_moves(self, data):
        self._moves = data

    def when_generate_is_called(self):
        session = {"nlg": self._data}
        self._result = nlg.generate(self._moves, self._context, session, self.logger)

    def then_result_contains(self, expected):
        for key in expected:
            assert expected[key] == self._result[key]

    def then_result_contains_one_of(self, possible_results):
        def entry_lists_match(expected, actual):
            for i in range(0, len(expected)):
                if not entries_match(expected[i], actual):
                    print("no match:", expected[i], actual)
                    return False
            return True

        def entries_match(expected, actual):
            return expected.get("uttrance", None) == actual.get("uttrance", None) \
                and expected.get("persona", None) == actual.get("persona", None)

        assert any([entry_lists_match(expected, self._result) for expected in possible_results])

    def test_basic_request(self):
        self.given_moves(["greet"])
        self.when_generate_is_called()
        self.then_result_contains_one_of([[{
            "utterance": "Hello!",
            "persona": None,
            "voice": None
        }], [{
            "utterance": "Hi!",
            "persona": None,
            "voice": None
        }]])

    def test_two_moves(self):
        self.given_moves(["icm:acc*pos", "ask(?X.city_mock_uuid_1(X))"])
        self.when_generate_is_called()
        self.then_result_contains({'persona': 'tutor', 'utterance': 'Ok. What city do you want the temperature for?'})

    def test_get_utterance_of_two_moves(self):
        self.given_moves(["icm:acc*pos", "ask(?X.city_mock_uuid_1(X))"])
        self.when_generate_utterance_is_called()
        self.then_result_is('Ok. What city do you want the temperature for?')

    def when_generate_utterance_is_called(self):
        session = {"nlg": self._data}
        self._result = nlg.generate_utterance(self._moves, {}, session, self.logger)

    def then_result_is(self, result):
        assert result == self._result

    def test_generalized_slots_for_grammar_entries_in_db(self):
        self.given_facts({
            "temperature_mock_uuid_0": {
                "sort": "some_sort",
                "value": "20_mock_uuid_5",
            },
            "city_mock_uuid_1": {
                "sort": "other_sort",
                "value": "gothenburg_mock_uuid_2",
            }
        })
        self.given_moves(["answer(temperature_mock_uuid_0(20_mock_uuid_5))"])
        self.when_generate_is_called()
        self.then_result_contains({
            'persona': "tutor",
            'utterance': 'The temperature in Gothenburg is 20 degrees centigrade.'
        })

    def test_generalized_slots_for_grammar_entries_in_db_dont_care_about_individual(self):
        self.given_facts({
            "temperature_mock_uuid_0": {
                "sort": "some_sort",
                "value": "20_mock_uuid_5",
            },
            "city_mock_uuid_1": {
                "sort": "other_sort",
                "value": "gothenburg_mock_uuid_2",
            }
        })
        self.given_moves(["answer(temperature_mock_uuid_0(some_random_string))"])
        self.when_generate_is_called()
        self.then_result_contains({
            'persona': "tutor",
            'utterance': 'The temperature in Gothenburg is 20 degrees centigrade.'
        })

    def test_generalized_slots_from_facts(self):
        self.given_facts({
            "temperature_mock_uuid_0": {
                "sort": "some_sort",
                "value": "id_24333u4u",
                "grammar_entry": "30"
            },
            "city_mock_uuid_1": {
                "sort": "other_sort",
                "value": "id_24333u4u",
                "grammar_entry": "Paris"
            }
        })
        self.given_moves(["answer(temperature_mock_uuid_0(30_mock_uuid_5))"])
        self.when_generate_is_called()
        self.then_result_contains({
            'persona': "tutor",
            'utterance': 'The temperature in Paris is 30 degrees centigrade.'
        })

    def test_grammar_entry_overrides_other_data(self):
        self.given_facts({
            "temperature_mock_uuid_0": {
                "sort": "some_sort",
                "value": "3000_mock_uuid_10",
            },
            "city_mock_uuid_1": {
                "sort": "other_sort",
                "value": "hell",
            }
        })
        self.given_moves(["answer(temperature_mock_uuid_0(3000_mock_uuid_10))"])
        self.when_generate_is_called()
        self.then_result_contains({
            'persona': "tutor",
            'utterance': "The temperature in hell can't be measured in centigrades."
        })

    def test_generalized_slots_for_string_individuals(self):
        self.given_facts({
            "temperature_mock_uuid_0": {
                "sort": "some_sort",
                "value": "20_mock_uuid_5",
            },
            "city_mock_uuid_1": {
                "sort": "string",
                "value": "Barcelona"
            }
        })
        self.given_moves(["answer(temperature_mock_uuid_0(20_mock_uuid_5))"])
        self.when_generate_is_called()
        self.then_result_contains({
            'persona': "tutor",
            'utterance': 'The temperature in Barcelona is 20 degrees centigrade.'
        })



from tala.utils.as_json import AsJSONMixin
from tala.utils.json_api import JSONAPIMixin, JSONAPIObject


class UnexpectedParameterFieldException(Exception):
    pass


class UnexpectedActionException(Exception):
    pass


class UnexpectedQueryException(Exception):
    pass


class UnexpectedValidatorException(Exception):
    pass


class DuplicateNameException(Exception):
    pass


class FailureReasonsNotAllowedException(Exception):
    pass


class UnsupportedServiceInterfaceTarget(Exception):
    pass


class ParameterNotFoundException(Exception):
    pass


class ParameterField:
    VALUE = "value"
    GRAMMAR_ENTRY = "grammar_entry"


SEMANTIC_OBJECT_TYPE = "service_interface"

FRONTEND_TARGET = "frontend_target"
HTTP_TARGET = "http_target"


class ServiceInterface(AsJSONMixin, JSONAPIMixin):
    @classmethod
    def create_from_json_api_data(cls, json_dict, _included):

        actions = [
            ServiceActionInterface.create_from_json_api_dict(action)
            for action in json_dict["attributes"].get("actions", [])
        ]

        queries = [
            ServiceQueryInterface.create_from_json_api_dict(query)
            for query in json_dict["attributes"].get("queries", [])
        ]

        validators = [
            ServiceValidatorInterface.create_from_json_api_dict(validator)
            for validator in json_dict["attributes"].get("validators", [])
        ]

        return cls(actions, queries, validators)

    def __init__(self, actions, queries, validators):
        self._validate(actions)
        self._actions = {action.name: action for action in actions}
        self._validate(queries)
        self._queries = {query.name: query for query in queries}
        self._validate(validators)
        self._validators = {validator.name: validator for validator in validators}

    def as_json_api_dict(self):

        service_interface = JSONAPIObject("tala.ddd.services.ServiceInterface", "ServiceInterface")

        for _name, action in self._actions.items():
            action_dict = action.as_json_api_dict()
            service_interface.append_attribute("actions", action_dict)

        for _name, query in self._queries.items():
            query_dict = query.as_json_api_dict()
            service_interface.append_attribute("queries", query_dict)

        for _name, validator in self._validators.items():
            validator_dict = validator.as_json_api_dict()
            service_interface.append_attribute("validators", validator_dict)

        return service_interface.as_dict

    def _validate(self, specific_interfaces):
        names = [interface.name for interface in specific_interfaces]
        if not self._all_unique(names):
            raise DuplicateNameException(
                "Expected all names to be unique among %s but they weren't" % specific_interfaces
            )

    def _all_unique(self, all_names):
        unique_names = set(all_names)
        return len(all_names) == len(unique_names)

    @property
    def actions(self):
        return list(self._actions.values())

    def get_action(self, name):
        if not self.has_action(name):
            raise UnexpectedActionException(
                "Expected one of the known actions %s but got '%s'" % (list(self._actions.keys()), name)
            )
        return self._actions[name]

    def has_action(self, name):
        return name in self._actions

    def get_query(self, name):
        if not self.has_query(name):
            raise UnexpectedQueryException(
                "Expected one of the known queries %s but got '%s'" % (list(self._queries.keys()), name)
            )
        return self._queries[name]

    def has_query(self, name):
        return name in self._queries

    def get_validator(self, name):
        if not self.has_validator(name):
            raise UnexpectedValidatorException(
                "Expected one of the known validators %s but got '%s'" % (list(self._validators.keys()), name)
            )
        return self._validators[name]

    def has_validator(self, name):
        return name in self._validators

    @property
    def queries(self):
        return list(self._queries.values())

    @property
    def validators(self):
        return list(self._validators.values())

    def __repr__(self):
        return "%s(actions=%s, queries=%s, validators=%s)" % (
            self.__class__.__name__, self.actions, self.queries, self.validators
        )

    def __eq__(self, other):
        def has_all(these, those):
            return all(this in those for this in these) and all(that in these for that in those)

        return bool(
            isinstance(other, self.__class__) and has_all(self.actions, other.actions)
            and has_all(self.queries, other.queries) and has_all(self.validators, other.validators)
        )

    def as_dict(self):
        json = super(ServiceInterface, self).as_dict()
        json["semantic_object_type"] = SEMANTIC_OBJECT_TYPE
        json["actions"] = [value.as_json() for value in self._actions.values()]
        json["queries"] = [value.as_json() for value in self._queries.values()]
        json["validators"] = [value.as_json() for value in self._validators.values()]
        return json


class SpecificServiceInterface(AsJSONMixin, JSONAPIMixin):
    def __init__(self, interface_type, name, target):
        super(SpecificServiceInterface, self).__init__()
        self._interface_type = interface_type
        self._name = name
        self._target = target

    @property
    def interface_type(self):
        return self._interface_type

    @property
    def name(self):
        return self._name

    @property
    def target(self):
        return self._target

    def __repr__(self):
        return "%s(%r, %r, %r)" % (self.__class__.__name__, self.interface_type, self.name, self.target)

    def __eq__(self, other):
        return bool(
            isinstance(other, self.__class__) and self.interface_type == other.interface_type
            and self.name == other.name and self.target == other.target
        )

    def ensure_target_is_not_frontend(self):
        if self.target.is_frontend:
            raise UnsupportedServiceInterfaceTarget(
                "Expected a non-frontend target for service interface '%s' but got a frontend target." % self.name
            )


class ParameterizedSpecificServiceInterface(SpecificServiceInterface):
    def __init__(self, interface_type, name, target, parameters):
        super(ParameterizedSpecificServiceInterface, self).__init__(interface_type, name, target)
        self._parameters = parameters

    @classmethod
    def create_parameters_from_json_api_dict(cls, data):
        name = data["id"]
        attributes = data["attributes"]
        if attributes["target"]["target_type"] == HTTP_TARGET:
            target = HttpTarget(attributes["target"]["endpoint"])
        elif attributes["target"]["target_type"] == FRONTEND_TARGET:
            target = FrontendTarget()
        parameters = [ServiceParameter.create_from_json_api_dict(parameter) for parameter in attributes["parameters"]]

        return name, target, parameters

    @property
    def parameters(self):
        return self._parameters

    def __repr__(self):
        return "%s(%r, %r, parameters=%s)" % (self.__class__.__name__, self.name, self.target, self.parameters)

    def __eq__(self, other):
        return bool(
            isinstance(other, self.__class__) and self.name == other.name and self.target == other.target
            and self.parameters == other.parameters
        )


class ServiceActionInterface(ParameterizedSpecificServiceInterface):
    def __init__(self, name, target, parameters, failure_reasons):
        super(ServiceActionInterface, self).__init__("action", name, target, parameters)
        self._failure_reasons = failure_reasons
        self._validate_target_and_failure_reasons()

    def _validate_target_and_failure_reasons(self):
        if self.target.is_frontend:
            if self.failure_reasons:
                failure_reason_names = [reason.name for reason in self.failure_reasons]
                raise FailureReasonsNotAllowedException(
                    "Expected no failure reasons for action '%s' with target 'frontend', but got %s" %
                    (self.name, failure_reason_names)
                )

    @classmethod
    def create_from_json_api_dict(cls, data):
        parameters = super().create_parameters_from_json_api_dict(data)
        failure_reasons = [ActionFailureReason(reason) for reason in data["attributes"]["failure_reasons"]]
        parameters += (failure_reasons, )
        return cls(*parameters)

    @property
    def failure_reasons(self):
        return self._failure_reasons

    def __repr__(self):
        return "%s(%r, %r, parameters=%s, failure_reasons=%s)" % (
            self.__class__.__name__, self.name, self.target, self.parameters, self.failure_reasons
        )

    def __eq__(self, other):
        return bool(
            isinstance(other, self.__class__) and self.name == other.name and self.target == other.target
            and self.parameters == other.parameters and self.failure_reasons == other.failure_reasons
        )

    def as_json_api_dict(self):
        return {
            "type": "tala.ddd.services.service_interface.ServiceActionInterface",
            "id": self.name,
            "attributes": {
                "target": self.target.as_json_api_attribute(),
                "parameters": [parameter.as_json_api_attribute() for parameter in self.parameters],
                "failure_reasons": [reason.name for reason in self.failure_reasons]
            }
        }


class ServiceQueryInterface(ParameterizedSpecificServiceInterface):
    def __init__(self, *args, **kwargs):
        super(ServiceQueryInterface, self).__init__("query", *args, **kwargs)
        self.ensure_target_is_not_frontend()

    @classmethod
    def create_from_json_api_dict(cls, data):
        parameters = super().create_parameters_from_json_api_dict(data)
        return cls(*parameters)

    def as_json_api_dict(self):
        return {
            "type": "tala.ddd.services.service_interface.ServiceQueryInterface",
            "id": self.name,
            "attributes": {
                "target": self.target.as_json_api_attribute(),
                "parameters": [parameter.as_json_api_attribute() for parameter in self.parameters],
            }
        }


class ServiceValidatorInterface(ParameterizedSpecificServiceInterface):
    def __init__(self, *args, **kwargs):
        super(ServiceValidatorInterface, self).__init__("validator", *args, **kwargs)
        self.ensure_target_is_not_frontend()

    def as_json_api_dict(self):
        return {
            "type": "tala.ddd.services.service_interface.ServiceValidatorInterface",
            "id": self.name,
            "attributes": {
                "target": self.target.as_json_api_attribute(),
                "parameters": [parameter.as_json_api_attribute() for parameter in self.parameters],
            }
        }

    @classmethod
    def create_from_json_api_dict(cls, data):
        parameters = super().create_parameters_from_json_api_dict(data)
        return cls(*parameters)


class ServiceImplicationInterface(SpecificServiceInterface):
    def __init__(self, *args, **kwargs):
        super(ServiceImplicationInterface, self).__init__("implication", *args, **kwargs)

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.name, self.target)


class ServiceParameter(AsJSONMixin):
    VALID_FORMATS = [ParameterField.VALUE, ParameterField.GRAMMAR_ENTRY]

    @classmethod
    def create_from_json_api_dict(cls, data):
        return cls(data["name"], data["format"], data["optional"])

    def __init__(self, name, format=None, is_optional=None):
        self._name = name
        is_optional = is_optional or False
        format = format or ParameterField.VALUE
        if format not in self.VALID_FORMATS:
            raise UnexpectedParameterFieldException(
                "Expected format as one of %s but got '%s' for parameter '%s'" % (self.VALID_FORMATS, format, name)
            )
        self._format = format
        self._is_optional = is_optional

    @property
    def name(self):
        return self._name

    @property
    def format(self):
        return self._format

    def as_json_api_attribute(self):
        return {"name": self.name, "format": self.format, "optional": self.is_optional}

    @property
    def is_optional(self):
        return self._is_optional

    def __repr__(self):
        return "%s(%r, format=%r, is_optional=%r)" % (self.__class__.__name__, self.name, self.format, self.is_optional)

    def __eq__(self, other):
        return bool(
            isinstance(other, self.__class__) and self.name == other.name and self.format == other.format
            and self.is_optional == other.is_optional
        )


class ActionFailureReason(AsJSONMixin):
    def __init__(self, name):
        super(ActionFailureReason, self).__init__()
        self._name = name

    @property
    def name(self):
        return self._name

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.name)

    def __eq__(self, other):
        return bool(isinstance(other, self.__class__) and self.name == other.name)


class ServiceTarget(AsJSONMixin):
    def __init__(self, target_type):
        self.target_type = target_type

    @property
    def is_frontend(self):
        return self.target_type == FRONTEND_TARGET

    @property
    def is_http(self):
        return self.target_type == HTTP_TARGET

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    def __eq__(self, other):
        return isinstance(other, self.__class__)

    def __ne__(self, other):
        return not (self == other)


class FrontendTarget(ServiceTarget):
    def __init__(self):
        super(FrontendTarget, self).__init__(FRONTEND_TARGET)

    def as_json_api_attribute(self):
        return {"target_type": FRONTEND_TARGET}


class HttpTarget(ServiceTarget):
    def __init__(self, endpoint):
        super(HttpTarget, self).__init__(HTTP_TARGET)
        self._endpoint = endpoint

    @property
    def endpoint(self):
        return self._endpoint

    def as_json_api_attribute(self):
        return {"target_type": HTTP_TARGET, "endpoint": self.endpoint}

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.endpoint)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.endpoint == other.endpoint



import warnings

from tala.utils.as_json import AsJSONMixin
from tala.utils.unicodify import unicodify


class StackError(Exception):
    pass


class Stack(AsJSONMixin):
    def __init__(self, content=None):
        if content is None:
            content = []
        super(Stack, self).__init__()
        self.content = list()
        for x in content:
            self.push(x)

    def as_dict(self):
        return {
            "stack": self.content,
        }

    def __repr__(self):
        return "{name}(content={content})".format(name=self.__class__.__name__, content=self.content)

    def __str__(self):
        string = "Stack(" + unicodify(self.content) + ")"
        return string

    def __eq__(self, other):
        try:
            return self.content == other.content
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.__class__.__name__) + hash(self.content)

    def __ne__(self, other):
        return not (self == other)

    def push(self, element):
        self.content.insert(0, element)

    def push_stack(self, other_stack):
        other_stack_elements = list(other_stack)
        other_stack_elements.reverse()
        for element in other_stack_elements:
            self.push(element)

    def top(self):
        if len(self) < 1:
            raise StackError("Cannot call 'top()' when stacksize <= 0")
        return self.content[0]

    def is_top(self, element):
        try:
            return element == self.top()
        except StackError:
            return False

    def isTop(self, element):
        warnings.warn("Stack.isTop() is deprecated. Use Stack.is_top() instead.", DeprecationWarning, stacklevel=2)
        return self.is_top(element)

    def pop(self):
        if len(self) < 1:
            raise StackError("Cannot call 'pop()' when stacksize <= 0")
        return self.content.pop(0)

    def __len__(self):
        return len(self.content)

    def isEmpty(self):
        warnings.warn("Stack.isEmpty() is deprecated. Use Stack.is_empty() instead.", DeprecationWarning, stacklevel=2)
        return self.is_empty()

    def is_empty(self):
        return len(self) == 0

    def clear(self):
        self.content = list()

    def remove(self, element):
        self.content.remove(element)

    def __iter__(self):
        return self.content.__iter__()


class StackSet(Stack):
    def __init__(self, content=None):
        if content is None:
            content = []
        super().__init__(content)

    def as_dict(self):
        return {
            "stackset": self.content,
        }

    def __str__(self):
        string = "stackset(" + unicodify(self.content) + ")"
        return string

    def push(self, element):
        if element in self.content:
            self.content.remove(element)
        self.content.insert(0, element)

    def remove_if_exists(self, element):
        if element in self.content:
            self.remove(element)

    def create_view(self, philter):
        return StackSetView(self, philter)


class StackSetView:
    def __init__(self, source_object, philter):
        self.source_object = source_object
        self.philter = philter

    def __str__(self):
        string = "stacksetview(" + str(list(self)) + ")"
        return string

    def top(self):
        return self._filtered_top(self.philter)

    def _filtered_top(self, philter):
        for element in self.source_object:
            if philter(element):
                return element

    def push(self, element):
        if element != self.top():
            self.source_object.push(element)

    def pop(self):
        self.source_object.remove(self.top())

    def __iter__(self):
        for elem in self.source_object:
            if self.philter(elem):
                yield elem

    def __len__(self):
        return len(list(self.__iter__()))

    def is_empty(self):
        return len(self) == 0



import structlog
import uuid

import pytest

from tala.utils.mqtt_client import MQTTClient
from tala.utils.func import configure_stdout_logging, getenv

logger = structlog.get_logger(__name__)
log_level = getenv("LOG_LEVEL", default="DEBUG")
configure_stdout_logging(log_level)


class TestMQTTClient:
    def setup_method(self):
        pass

    def test_creation(self):
        self.given_args_for_client_creation(["", logger, "some_endpoint", 443])
        self.when_mqtt_client_created()
        self.then_client_created_with(logger, "some_endpoint", 443, "name_base")

    def given_args_for_client_creation(self, args):
        self._mqtt_args = args

    def when_mqtt_client_created(self):
        self._mqtt_client = MQTTClient(*self._mqtt_args)

    def then_client_created_with(self, logger, endpoint, port, name_base):
        assert self._mqtt_client.logger == logger
        assert self._mqtt_client._endpoint == endpoint
        assert self._mqtt_client._port == port

    def test_creation_with_client_id(self):
        self.given_args_for_client_creation(["client_id", logger, "some_endpoint", 443])
        self.when_mqtt_client_created()
        self.then_client_id_matches("client_id")

    def then_client_id_matches(self, id_):
        assert self._mqtt_client._client_id.startswith(f"{id_}-")

    @pytest.mark.parametrize("chunk", ["This is an unproblematic test utterance.", "Frölunda ", "Fr\u00f6lunda"])
    def test_stream_single_chunk(self, chunk):
        self.given_mqtt_client_started("test_client", logger, "tala-sse.azurewebsites.net", 443)
        self.given_session_id("test-session-id")
        self.when_chunk_streamed(chunk)
        self.then_everything_is_ok()

    def given_mqtt_client_started(self, name, logger, endpoint, port):
        self._client = MQTTClient(name, logger, endpoint, port)
        self._client.start()

    def given_session_id(self, id_):
        self._session_id = id_ + str(uuid.uuid4)

    def when_chunk_streamed(self, chunk):
        self._client.open_session(self._session_id, None)
        self._client.set_persona(self._session_id, "some-persona")
        self._client.set_voice(self._session_id, "some-voice")
        self._client.stream_chunk(self._session_id, chunk)
        self._client.flush_stream(self._session_id)
        self._client.close_session(self._session_id)

    def then_everything_is_ok(self):
        assert True
